# -*- coding: utf-8 -*-

# Copyright Daniel Gonzalez     <dgzaballos@gmail.com>
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os.path

from gluon import current, DAL, HTTP
from gluon.tools import Auth
from gluon.utils import web2py_uuid

from system.environment import Environment

from core.authorization import OMWAuth

class Application(object):
    def __init__(self, app_path, config):
        self.app_path = app_path
        
        self.config = self._configuration(config)
        self.db = self._dal()        
        self.auth = self._get_auth()
        self.environment = self._get_environment()
        
        self._set_superadmin()
        
    def _configuration(self, config):
        # Start configuration system        
        if config.has('hmac_key') is False:            
            config.set('hmac_key', web2py_uuid())
            
        return config
            
    def _dal(self):
        migrate_param = self.config.get('dal_migrate')
        if migrate_param is None:
            migrate_param = True
        return DAL(self.config.get('dal'), migrate=migrate_param)
    
    def _get_auth(self):            
        ## Enable application as CAS consumer
        if self.config.has('cas_provider'): 
            auth = OMWAuth(db=self.db, 
                           hmac_key=self.config.get('hmac_key'), 
                           mailer=self.config.get('email_active'), 
                           cas_provider=self.config.get('cas_provider'))
        else:
            auth = OMWAuth(db=self.db, 
                           hmac_key=self.config.get('hmac_key'), 
                           mailer=self.config.get('email_active')) 
            
            ## Enable application as CAS provider
            if self.config.has('cas_domain'):             
                auth.settings.cas_domains.append(self.config.get('cas_domain'))            
        
        ## configure auth policy
        auth_config = self.config.get('register')
        auth.settings.password_min_length = auth_config['min_password_length']
        auth.settings.registration_requires_verification =  auth_config['registration_verification']
        auth.settings.registration_requires_approval =  auth_config['registration_approval']
        auth.settings.reset_password_requires_verification =  auth_config['reset_password_verification']
            
        if self.config.get('actions_disabled') is None:
            auth.settings.actions_disabled = ['profile', 'change_password',
                                              'register', 'retrieve_username', 
                                              'request_reset_password']
        else:
            auth.settings.actions_disabled = self.config.get('actions_disabled')
        return auth
    
    def _get_environment(self):
        environ = Environment()
        if current.session.environ:
            environ.set(current.session.environ)
        else:
            environ.set()
        return environ
    
    def _set_superadmin(self):
        # Add superadmin group and user
        if not self.config.get('admin_create'):
            # Superadmin group
            admin_id = self.auth.add_group('admin','Application management')
            
            from masters.users import Users
            users = Users(self.config, self.db)
            admin_user = self.config.get('admin')
            # Needed into htpasswd file process
            current._config = self.config
            user = users.add(admin_user['first_name'],
                             admin_user['last_name'],
                             admin_user['email'],
                             admin_user['password'],
                             False) # Do not notify
            if user is None:
                current.response.flash = user.errors
            else:
                # Add admin user to admin group
                self.auth.add_membership(admin_id, user )
                self.config.set('admin_create', True)
                
    def set_mailer(self):
        ## configure email
        mail = None
        if self.config.get('email_active'):     
            mail = self.auth.settings.mailer
            mail.settings.server = self.config.get('email_server')
            mail.settings.sender = self.config.get('email_sender')
            mail.settings.login = self.config.get('email_login')
            
        return mail