# -*- coding: utf-8 -*-

# Copyright Daniel Gonzalez     <dgzaballos@gmail.com>
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os
import yaml

from mil.base import YAMLFile

from gluon import current
from gluon.storage import Storage

class GovernmentHandler():
    @staticmethod
    def get_government_files():
        # 
        g_path = current._config.government_path
        files = []
        for f in os.listdir(g_path):
            if f not in ['.svn']:
                items = f[0:-5].split('_')
                files.append({'app_name': items[1],
                              'version': items[2],
                              'path': f[0:-5]})
        
        return files
    
    @staticmethod
    def process_government_file(file):
        path = os.path.join(current._config.government_path, '%s.yaml' % file)
        
        if path:
            GovernmentHandler.__prepare_conf(YAMLFile.load(path))
        else:
            lit = current.T("Government file '%s' not exists")
            raise Exception(lit % file)             
         
    @staticmethod   
    def __prepare_conf(data):        
        # Process data
        if data is not None:        
            try:
                # Get metadata
                metadata = data['metadata']
                application = metadata['application']
                
                # Checks if application is in environment
                if not application in current._config.get('subapp'):
                    lit =  current.T('Application "%s" not in environment')
                    raise Exception(lit % application)
                                
                from system import AuthConfiguration
                
                # Process Configuration
                configuration = data['configuration']
                
                if configuration:
                    auth_mng = AuthConfiguration(current.db)     # Role == 0
                    
                    # Clean application data
                    auth_mng.delete(application)       # Module == application
                    
                    for item in configuration:
                        # Add new row (update or insert)
                        auth_mng.add(application,
                                     item['module'], 
                                     item['action'], 
                                     item['value'])        
                
                # Process Government Profiles
                profiles = data['government']['profiles']
                
                if profiles:
                    # Import model
                    from masters.profiles import Profiles
                    insKlss = Profiles(current._config, current.db)
                    
                    profiles_id = {}
                    for item in profiles:
                        # Test if role exists
                        pid = insKlss.get_id_by_subapp_and_role(application, 
                                                                item['role'])
                        
                        if pid is not None:         
                            profiles_id[item['role']] = pid
                        else:                        
                            # Add web2py group
                            gp_id = current.auth.add_group(item['role'], 
                                                           item['description'])
                            
                            # Add profile to subapplication
                            pid = insKlss.add_profile(application, gp_id, True)
                            profiles_id[item['role']] = gp_id
    
    
                # Process Government Authorization
                permissions = data['government']['permissions']
                if permissions:
                    for role in permissions:
                        role_data = permissions[role]
                        auth_mng = AuthConfiguration(current.db, 
                                                     profiles_id[role])
                        for item in role_data:
                            # Add new row (update or insert)
                            auth_mng.add(application,
                                         item['module'], 
                                         item['action'], 
                                         item['value'])        
                    
            except Exception, e:
                raise e        
        else:
            raise current.T('Configuration file is empty') 
    
    @staticmethod
    def get_government_file(application):        
        if application:
            return GovernmentHandler.__get_conf(application)
        else:
            lit = current.T("Government file '%s' not exists")
            raise Exception(lit % application)
        
    @staticmethod   
    def __get_conf(application):        
        # Process data
        if application is not None:        
            try:
                data = {'metadata': {'application': application},
                        'configuration': [],
                        'government': {'profiles': [{'role': 'leader', 
                                                     'description': 'Leader'},
                                                    {'role': 'collaborator',
                                                     'description': 'Collaborator'},
                                                    {'role': 'guest',
                                                     'description': 'Guest'},
                                                    {'role': 'public',
                                                     'description': 'Public'}],
                                       'permissions': {}}}
                
                # Configuration data
                from system import AuthConfiguration
                auth = AuthConfiguration(current.db)
                
                # Config
                rows = auth.get_by_application_and_role(application, 0)
                
                for row in rows:
                    data['configuration'].append({'module': row.module,
                                                  'action': row.action,
                                                  'value': row.value})
                # Profiles
                from masters.profiles import Profiles
                insKlss = Profiles(current._config, current.db)
                
                for item in data['government']['profiles']:
                    # Test if role exists
                    pid = insKlss.get_id_by_subapp_and_role(application, 
                                                            item['role'])
                    
                    rows = auth.get_by_application_and_role(application, pid)
                    
                    tmp = []
                    for row in rows:
                        tmp.append({'module': row.module,
                                    'action': row.action,
                                    'value': row.value})
                    data['government']['permissions'][item['role']] = tmp
                        
                
                return yaml.dump(data, default_flow_style=False)
            except Exception, e:
                raise e        
        else:
            raise current.T('Configuration file is empty')          