# -*- coding: utf-8 -*-
#
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import hashlib
import mimetypes
import re
import time
from datetime import datetime

class BaseDriver(object):
    """Dummy driver for elFinder"""
    
    options = {}
    
    time = 0
    request = {}
    response = {}
    errorData = {}
    form = {}
    im = None
    sp = None
    today = 0
    yesterday = 0

    # public variables
    httpAllowedParameters = ('cmd', 'target', 'targets[]', 'current', 'tree', 
                             'name', 'content', 'src', 'dst', 'cut', 'init', 
                             'type', 'width', 'height', 'upload[]')
    
    commands = {
        'open':    '__open',
        'reload': '__reload',
        'mkdir': '__mkdir',
        'mkfile': '__mkfile',
        'rename': '__rename',
        'upload': '__upload',
        'paste': '__paste',
        'rm': '__rm',
        'duplicate': '__duplicate',
        'read': '__read',
        'edit': '__edit',
        'extract': '__extract',
        'archive': '__archive',
        'resize': '__resize',
        'tmb': '__thumbnails',
        'ping': '__ping'
    }
    
    mimeType = {
        # text
        'txt': 'text/plain',
        'conf': 'text/plain',
        'ini': 'text/plain',
        'php': 'text/x-php',
        'html': 'text/html',
        'htm': 'text/html',
        'js' : 'text/javascript',
        'css': 'text/css',
        'rtf': 'text/rtf',
        'rtfd': 'text/rtfd',
        'py' : 'text/x-python',
        'java': 'text/x-java-source',
        'rb' : 'text/x-ruby',
        'sh' : 'text/x-shellscript',
        'pl' : 'text/x-perl',
        'sql': 'text/x-sql',
        # apps
        'doc': 'application/msword',
        'ogg': 'application/ogg',
        '7z': 'application/x-7z-compressed',
        # video
        'ogm': 'appllication/ogm',
        'mkv': 'video/x-matroska'
    }

    # return variables
    httpStatusCode = 0
    httpHeader = {}
    httpResponse = None
    
    def __init__(self, opts):
        for opt in opts:
            self.options[opt] = opts.get(opt)            
        
        self.response['debug'] = {}
        
        for cmd in self.options['disabled']:
            if cmd in self.commands:
                del self.commands[cmd]
        self.__clean()
                
    def __clean(self):
        self.httpStatusCode = 0
        self.httpHeader = {}
        self.httpResponse = None
          
    def reset(self):
        """Flush per request variables"""
        self.request = {}
        self.response = {}
        self.errorData = {}
        self.form = {}

        self.time = time.time()
        t = datetime.fromtimestamp(self.time)
        self.today = time.mktime(datetime(t.year, t.month, t.day).timetuple())
        self.yesterday = self.today - 86400

        self.response['debug'] = {}
        
    def run(self, httpRequest=[]):
        self._main_process(httpRequest)
                
        if self.errorData:
            self.response['errorData'] = self.errorData

        if self.options['debug']:
            self.debug('time', (time.time() - self.time))
        else:
            if 'debug' in self.response:
                del self.response['debug']
                
        if self.httpStatusCode < 100:
            self.httpStatusCode = 200

        if not 'Content-type' in self.httpHeader:
            if ('cmd' in self.request and self.request['cmd'] == 'upload') or self.options['debug']:
                self.httpHeader['Content-type'] = 'text/html'
            else:
                self.httpHeader['Content-type'] = 'application/json'
        
        self.httpResponse = self.response
                
        return (self.httpStatusCode, self.httpHeader, self.httpResponse)


    # Class API
    def debug(self, k, v):
        if self.options['debug']:
            self.response['debug'].update({k: v})
        return

    def get_content(self, path, tree):
        """CWD + CDC + maybe(TREE)"""
        self.response['cwd'] = None
        self.response['cdc'] = None

        if tree:
            self.response['tree']  = None

    def check_name(self, name):
        """Check for valid file/dir name"""
        pattern = r'[\/\\\:\<\>]'
        if re.search(pattern, name):
            return False
        return True
    
    def crop_tuple(self, size):
        w, h = size
        if w > h: # landscape
            l = int((w - h) / 2)
            u = 0
            r = l + h
            d = h
            return (l, u, r, d)
        elif h > w: # portrait
            l = 0
            u = int((h - w) / 2)
            r = w
            d = u + w
            return (l, u, r, d)
        else: # cube
            pass

        return False
    
    def is_upload_allow(self, name):
        allow = False
        deny = False
        mime = self.get_mimetype(name)

        if 'all' in self.options['uploadAllow']:
            allow = True
        else:
            for a in self.options['uploadAllow']:
                if mime.find(a) == 0:
                    allow = True

        if 'all' in self.options['uploadDeny']:
            deny = True
        else:
            for d in self.options['uploadDeny']:
                if mime.find(d) == 0:
                    deny = True

        if self.options['uploadOrder'][0] == 'allow': # ,deny
            if deny is True:
                return False
            elif allow is True:
                return True
            else:
                return False
        else: # deny,allow
            if allow is True:
                return True
            elif deny is True:
                return False
            else:
                return True


    def is_accepted(self, target):
        if target == '.' or target == '..':
            return False
        if target[0:1] == '.' and not self.options['dotFiles']:
            return False
        return True


    def get_hash(self, path):
        """Hash of the path"""
        m = hashlib.md5()
        m.update(path)
        return str(m.hexdigest())



    def get_mimetype(self, path):
        """Detect mimetype of file"""
        mime = mimetypes.guess_type(path)[0] or 'unknown'
        ext = path[path.rfind('.') + 1:]

        if mime == 'unknown' and ('.' + ext) in mimetypes.types_map:
            mime = mimetypes.types_map['.' + ext]

        if mime == 'text/plain' and ext == 'pl':
            mime = self.mimeType[ext]

        if mime == 'application/vnd.ms-office' and ext == 'doc':
            mime = self.mimeType[ext]

        if mime == 'unknown':
            if ext in self.mimeType:
                mime = self.mimeType[ext]

        return mime

    def get_img_lib(self):
        if not self.options['imgLib'] is False and self.im is None:
            try:
                #import Image
                from PIL.Image import Image
                Image
                self.im = Image
                self.options['imgLib'] = 'PIL'
            except:
                self.options['imgLib'] = False
                self.im = False

        self.debug('imgLib', self.options['imgLib'])
        return self.options['imgLib']


    def get_img_size(self, path):
        self.get_img_lib();
        if self.__canCreateTmb():
            try:
                im = self.im.open(path)
                return str(im.size[0]) + 'x' + str(im.size[1])
            except:
                pass

        return False

    def check_archivers(self):
        # import subprocess
        # sp = subprocess.Popen(['tar', '--version'], shell = False,
        # stdout = subprocess.PIPE, stderr=subprocess.PIPE)
        # out, err = sp.communicate()
        # print 'out:', out, '\nerr:', err, '\n'
        archive = { 'create': {}, 'extract': {} }
        c = archive['create']
        e = archive['extract']

        tar = self.run_sub_process(['tar', '--version'])
        gzip = self.run_sub_process(['gzip', '--version'])
        bzip2 = self.run_sub_process(['bzip2', '--version'])
        zipc = self.run_sub_process(['zip', '--version'])
        unzip = self.run_sub_process(['unzip', '--help'])
        rar = self.run_sub_process(['rar', '--version'], validReturn = [0, 7])
        unrar = self.run_sub_process(['unrar'], validReturn = [0, 7])
        p7z = self.run_sub_process(['7z', '--help'])
        p7za = self.run_sub_process(['7za', '--help'])
        p7zr = self.run_sub_process(['7zr', '--help'])

        # tar = False
        # tar = gzip = bzip2 = zipc = unzip = rar = unrar = False
        # print tar, gzip, bzip2, zipc, unzip, rar, unrar, p7z, p7za, p7zr

        if tar:
            mime = 'application/x-tar'
            c.update({mime: {'cmd': 'tar', 'argc': '-cf', 'ext': 'tar'}})
            e.update({mime: {'cmd': 'tar', 'argc': '-xf', 'ext': 'tar'}})

        if tar and gzip:
            mime = 'application/x-gzip'
            c.update({mime: {'cmd': 'tar', 'argc': '-czf', 'ext': 'tar.gz'}})
            e.update({mime: {'cmd': 'tar', 'argc': '-xzf', 'ext': 'tar.gz'}})

        if tar and bzip2:
            mime = 'application/x-bzip2'
            c.update({mime: {'cmd': 'tar', 'argc': '-cjf', 'ext': 'tar.bz2'}})
            e.update({mime: {'cmd': 'tar', 'argc': '-xjf', 'ext': 'tar.bz2'}})

        mime = 'application/zip'
        if zipc:
            c.update({mime: {'cmd': 'zip', 'argc': '-r9', 'ext': 'zip'}})
        if unzip:
            e.update({mime: {'cmd': 'unzip', 'argc': '', 'ext': 'zip'}})

        mime = 'application/x-rar'
        if rar:
            c.update({mime: {'cmd': 'rar', 'argc': 'a -inul', 'ext': 'rar'}})
            e.update({mime: {'cmd': 'rar', 'argc': 'x -y', 'ext': 'rar'}})
        elif unrar:
            e.update({mime: {'cmd': 'unrar', 'argc': 'x -y', 'ext': 'rar'}})

        p7zip = None
        if p7z:
            p7zip = '7z'
        elif p7za:
            p7zip = '7za'
        elif p7zr:
            p7zip = '7zr'

        if p7zip:
            mime = 'application/x-7z-compressed'
            c.update({mime: {'cmd': p7zip, 'argc': 'a -t7z', 'ext': '7z'}})
            e.update({mime: {'cmd': p7zip, 'argc': 'e -y', 'ext': '7z'}})

            mime = 'application/x-tar'
            if not mime in c:
                c.update({mime: {'cmd': p7zip, 'argc': 'a -ttar', 'ext': 'tar'}})
            if not mime in e:
                e.update({mime: {'cmd': p7zip, 'argc': 'e -y', 'ext': 'tar'}})

            mime = 'application/x-gzip'
            if not mime in c:
                c.update({mime: {'cmd': p7zip, 'argc': 'a -tgzip', 'ext': 'gz'}})
            if not mime in e:
                e.update({mime: {'cmd': p7zip, 'argc': 'e -y', 'ext': 'tar.gz'}})

            mime = 'application/x-bzip2'
            if not mime in c:
                c.update({mime: {'cmd': p7zip, 'argc': 'a -tbzip2', 'ext': 'bz2'}})
            if not mime in e:
                e.update({mime: {'cmd': p7zip, 'argc': 'e -y', 'ext': 'tar.bz2'}})

            mime = 'application/zip'
            if not mime in c:
                c.update({mime: {'cmd': p7zip, 'argc': 'a -tzip', 'ext': 'zip'}})
            if not mime in e:
                e.update({mime: {'cmd': p7zip, 'argc': 'e -y', 'ext': 'zip'}})

        if not self.options['archiveMimes']:
            self.options['archiveMimes'] = c.keys()
        else:
            pass

        self.options['archivers'] = archive
        pass


    def run_sub_process(self, cmd, validReturn = [0]):
        if self.sp is None:
            import subprocess
            self.sp = subprocess

        try:
            sp = self.sp.Popen(cmd, shell = False, stdout = self.sp.PIPE, stderr = self.sp.PIPE, stdin = self.sp.PIPE)
            out, err = sp.communicate('')
            ret = sp.returncode
            # print cmd, ret, out, err
        except:
            return False

        if not ret in validReturn:
            return False

        return True


    def check_utf8(self, name):
        try:
            name.decode('utf-8')
        except UnicodeDecodeError:
            name = unicode(name, 'utf-8', 'replace')
            self.debug('invalid encoding', name)
            #name += ' (invalid encoding)'
        return name


    def set_error(self, path, msg):
        """Collect error/warning messages"""
        self.errorData[path] = msg

