# -*- coding: utf-8 -*-
#
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os
import os.path
import re
import shutil
import time
from datetime import datetime

from gluon import current

from .base import BaseDriver

FileSystemOptions =  {
        'root': '',
        'URL': '',
        'rootAlias': 'Home',
        'dotFiles': False,
        'dirSize': True,
        'fileMode': 0644,
        'dirMode': 0755,
        'imgLib': 'auto',
        'tmbDir': '.tmb',
        'tmbAtOnce': 5,
        'tmbSize': 48,
        'fileURL': True,
        'uploadMaxSize': 256,
        'uploadWriteChunk': 8192,
        'uploadAllow': [],
        'uploadDeny': [],
        'uploadOrder': ['deny', 'allow'],
        # 'aclObj': None, # TODO
        # 'aclRole': 'user', # TODO
        'defaults': {
            'read': True,
            'write': True,
            'rm': True
        },
        'perms': {},
        'archiveMimes': {},
        'archivers': {},
        'disabled': [],
        'debug': False
    }

class FileSystemDriver(BaseDriver):
    """Connector for elFinder"""
    
    def __init__(self, opts):
        options = FileSystemOptions
        for opt in opts:
            options[opt] = opts.get(opt)          
        BaseDriver.__init__(self, options)

        self.options['URL'] = self.options['URL'].rstrip('/')
        self.options['root'] = self.options['root'].rstrip(os.sep)
        self.debug('URL', self.options['URL'])
        self.debug('root', self.options['root'])

        if self.options['tmbDir']:
            self.options['tmbDir'] = os.path.join(self.options['root'], 
                                                  self.options['tmbDir'])
            if not os.path.exists(self.options['tmbDir']):
                self.options['tmbDir'] = False

    def get_content(self, path, tree):
        """CWD + CDC + maybe(TREE)"""
        self.__cwd(path)
        self.__cdc(path)

        if tree:
            self.response['tree'] = self.__tree(self.options['root'])            

    def _main_process(self, httpRequest):
        rootOk = True
        if not os.path.exists(self.options['root']) or self.options['root'] == '':
            rootOk = False
            self.response['error'] = current.T('Invalid backend configuration')
        elif not self.__is_allowed(self.options['root'], 'read'):
            rootOk = False
            self.response['error'] = current.T('Access denied')

        for field in self.httpAllowedParameters:
            if field in httpRequest:
                self.request[field] = httpRequest[field]

        if rootOk is True:
            if 'cmd' in self.request:
                if self.request['cmd'] in self.commands:
                    cmd = self.commands[self.request['cmd']]
                    func = getattr(self, 
                                   '_' + self.__class__.__name__ + cmd, 
                                   None)
                    if callable(func):
                        try:
                            func()
                        except Exception, e:
                        	self.response['error'] = current.T('Command Failed')
                        	self.debug('exception', str(e))
                else:
                    self.response['error'] = current.T('Unknown command')
            else:
                self.__open()

            if 'init' in self.request:
                self.check_archivers()
                self.response['disabled'] = self.options['disabled']
                if not self.options['fileURL']:
                    url = ''
                else:
                    url = self.options['URL']
                self.response['params'] = {
                    'dotFiles': self.options['dotFiles'],
                    'uplMaxSize': str(self.options['uploadMaxSize']) + 'M',
                    'archives': self.options['archiveMimes'],
                    'extract': self.options['archivers']['extract'].keys(),
                    'url': url
                }
                
    def __open(self):
        """Open file or directory"""
        # try to open file
        if 'current' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            curFile = self.__find(self.request['target'], curDir)

            if not curDir or not curFile or os.path.isdir(curFile):
                self.httpStatusCode = 404
                self.httpHeader['Content-type'] = 'text/html'
                self.httpResponse = 'File not found'
                return
            if not self.__is_allowed(curDir, 'read') or \
               not self.__is_allowed(curFile, 'read'):
                self.httpStatusCode = 403
                self.httpHeader['Content-type'] = 'text/html'
                self.httpResponse = 'Access denied'
                return

            if os.path.islink(curFile):
                curFile = self.__readlink(curFile)
                if not curFile or os.path.isdir(curFile):
                    self.httpStatusCode = 404
                    self.httpHeader['Content-type'] = 'text/html'
                    self.httpResponse = 'File not found'
                    return
                if (
                    not self.__is_allowed(os.path.dirname(curFile), 'read')
                    or not self.__is_allowed(curFile, 'read')
                    ):
                    self.httpStatusCode = 403
                    self.httpHeader['Content-type'] = 'text/html'
                    self.httpResponse = 'Access denied'
                    return

            mime = self.get_mimetype(curFile)
            parts = mime.split('/', 2)
            if parts[0] == 'image': disp = 'image'
            elif parts[0] == 'text': disp = 'inline'
            else: disp = 'attachments'

            self.httpStatusCode = 200
            self.httpHeader['Content-type'] = mime
            self.httpHeader['Content-Disposition'] = \
                        disp + '; filename=' + os.path.basename(curFile)
            self.httpHeader['Content-Location'] = \
                        curFile.replace(self.options['root'], '')
            self.httpHeader['Content-Transfer-Encoding'] = 'binary'
            self.httpHeader['Content-Length'] = str(os.lstat(curFile).st_size)
            self.httpHeader['Connection'] = 'close'
            self.response['file'] = open(curFile, 'r')
            return
        # try dir
        else:
            path = self.options['root']

            if 'target' in self.request:
                target = self.__find_dir(self.request['target'], None)
                if not target:
                    self.response['error'] = current.T('Invalid parameters')
                elif not self.__is_allowed(target, 'read'):
                    self.response['error'] = current.T('Access denied')
                else:
                    path = target

            self.get_content(path, 'tree' in self.request)
        pass


    def __rename(self):
        """Rename file or dir"""
        current = name = target = None
        curDir = curName = newName = None
        if 'name' in self.request and \
           'current' in self.request and \
           'target' in self.request:
            name = self.request['name']
            current = self.request['current']
            target = self.request['target']
            curDir = self.__find_dir(current, None)
            curName = self.__find(target, curDir)
            newName = os.path.join(curDir, name)

        if not curDir or not curName:
            self.response['error'] = current.T('File not found')
        elif not self.__is_allowed(curDir, 'write') and \
             self.__is_allowed(curName, 'rm'):
            self.response['error'] = current.T('Access denied')
        elif not self.check_name(name):
            self.response['error'] = current.T('Invalid name')
        elif os.path.exists(newName):
            self.response['error'] = \
                    current.T('File or folder with the same name already exists')
        else:
            self.__rm_tmb(curName)
            try:
                os.rename(curName, newName)
                self.response['select'] = [self.get_hash(newName)]
                self.get_content(curDir, os.path.isdir(newName))
            except:
                self.response['error'] = current.T('Unable to rename file')


    def __mkdir(self):
        """Create new directory"""
        current = None
        path = None
        newDir = None
        if 'name' in self.request and 'current' in self.request:
            name = self.request['name']
            current = self.request['current']
            path = self.__find_dir(current, None)
            newDir = os.path.join(path, name)

        if not path:
            self.response['error'] = current.T('Invalid parameters')
        elif not self.__is_allowed(path, 'write'):
            self.response['error'] = current.T('Access denied')
        elif not self.check_name(name):
            self.response['error'] = current.T('Invalid name')
        elif os.path.exists(newDir):
            self.response['error'] = \
                    current.T('File or folder with the same name already exists')
        else:
            try:
                os.mkdir(newDir, int(self.options['dirMode']))
                self.response['select'] = [self.get_hash(newDir)]
                self.get_content(path, True)
            except:
                self.response['error'] = current.T('Unable to create folder')


    def __mkfile(self):
        """Create new file"""
        name = current = None
        curDir = newFile = None
        if 'name' in self.request and 'current' in self.request:
            name = self.request['name']
            current = self.request['current']
            curDir = self.__find_dir(current, None)
            newFile = os.path.join(curDir, name)

        if not curDir or not name:
            self.response['error'] = current.T('Invalid parameters')
        elif not self.__is_allowed(curDir, 'write'):
            self.response['error'] = current.T('Access denied')
        elif not self.check_name(name):
            self.response['error'] = current.T('Invalid name')
        elif os.path.exists(newFile):
            self.response['error'] = \
                    current.T('File or folder with the same name already exists')
        else:
            try:
                open(newFile, 'w').close()
                self.response['select'] = [self.get_hash(newFile)]
                self.get_content(curDir, False)
            except:
                self.response['error'] = current.T('Unable to create file')


    def __rm(self):
        """Delete files and directories"""
        current = rmList = None
        curDir = rmFile = None
        if 'current' in self.request and 'targets[]' in self.request:
            current = self.request['current']
            rmList = self.request['targets[]']
            curDir = self.__find_dir(current, None)

        if not rmList or not curDir:
            self.response['error'] = current.T('Invalid parameters')
            return False

        if not isinstance(rmList, list):
            rmList = [rmList]

        for rm in rmList:
            rmFile = self.__find(rm, curDir)
            if not rmFile: continue
            self.__remove(rmFile)
        # TODO if errorData not empty return error
        self.get_content(curDir, True)


    def __upload(self):
        """Upload files"""
        try: # Windows needs stdio set for binary mode.
            import msvcrt
            msvcrt.setmode (0, os.O_BINARY) # stdin  = 0
            msvcrt.setmode (1, os.O_BINARY) # stdout = 1
        except ImportError:
            pass

        if 'current' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            if not curDir:
                self.response['error'] = current.T('Invalid parameters')
                return
            if not self.__is_allowed(curDir, 'write'):
                self.response['error'] = current.T('Access denied')
                return
            if not 'upload[]' in self.request:
                self.response['error'] = current.T('No file to upload')
                return

            upFiles = self.request['upload[]']
            # invalid format
            # must be dict('filename1': 'filedescriptor1', 
            #              'filename2': 'filedescriptor2', ...)
            if not isinstance(upFiles, dict):
                self.response['error'] = current.T('Invalid parameters')
                return

            self.response['select'] = []
            total = 0
            upSize = 0
            maxSize = self.options['uploadMaxSize'] * 1024 * 1024
            for name, data in upFiles.iteritems():
                if name:
                    total += 1
                    name = os.path.basename(name)
                    if not self.check_name(name):
                        self.set_error(name, 'Invalid name')
                    else:
                        name = os.path.join(curDir, name)
                        try:
                            f = open(name, 
                                     'wb', 
                                     self.options['uploadWriteChunk'])
                            for chunk in self.__fbuffer(data):
                                f.write(chunk)
                            f.close()
                            upSize += os.lstat(name).st_size
                            if self.is_upload_allow(name):
                                os.chmod(name, self.options['fileMode'])
                                self.response['select'].append(self.get_hash(name))
                            else:
                                self.set_error(name, 'Not allowed file type')
                                try:
                                    os.unlink(name)
                                except:
                                    pass
                        except:
                            self.set_error(name, 'Unable to save uploaded file')
                        if upSize > maxSize:
                            try:
                                os.unlink(name)
                                self.set_error(name, 
                                   'File exceeds the maximum allowed filesize')
                            except:
                                pass
                                # TODO ?
                                self.set_error(name, 
                                            'File was only partially uploaded')
                            break

            if self.errorData:
                if len(self.errorData) == total:
                    self.response['error'] = current.T('Unable to upload files')
                else:
                    self.response['error'] = current.T('Some files was not uploaded')

            self.get_content(curDir, False)
            return


    def __paste(self):
        """Copy or cut files/directories"""
        if 'current' in self.request and \
           'src' in self.request and \
           'dst' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            src = self.__find_dir(self.request['src'], None)
            dst = self.__find_dir(self.request['dst'], None)
            if not curDir or not src or not dst or \
               not 'targets[]' in self.request:
                self.response['error'] = current.T('Invalid parameters')
                return
            files = self.request['targets[]']
            if not isinstance(files, list):
                files = [files]

            cut = False
            if 'cut' in self.request:
                if self.request['cut'] == '1':
                    cut = True

            if not self.__is_allowed(src, 'read') or \
               not self.__is_allowed(dst, 'write'):
                self.response['error'] = current.T('Access denied')
                return

            for fhash in files:
                f = self.__find(fhash, src)
                if not f:
                    self.response['error'] = current.T('File not found')
                    return
                newDst = os.path.join(dst, os.path.basename(f))
                if dst.find(f) == 0:
                    self.response['error'] = current.T('Unable to copy into itself')
                    return

                if cut:
                    if not self.__is_allowed(f, 'rm'):
                        self.response['error'] = current.T('Move failed')
                        self.errorData(f, 'Access denied')
                        self.get_content(curDir, True)
                        return
                    # TODO thumbs
                    if os.path.exists(newDst):
                        self.response['error'] = current.T('Unable to move files')
                        self.errorData(f, 
                            'File or folder with the same name already exists')
                        self.get_content(curDir, True)
                        return
                    try:
                        os.rename(f, newDst)
                        self.__rm_tmb(f)
                        continue
                    except:
                        self.response['error'] = current.T('Unable to move files')
                        self.errorData(f, 'Unable to move')
                        self.get_content(curDir, True)
                        return
                else:
                    if not self.__copy(f, newDst):
                        self.response['error'] = current.T('Unable to copy files')
                        self.get_content(curDir, True)
                        return
                    continue
            self.get_content(curDir, True)
        else:
            self.response['error'] = current.T('Invalid parameters')

        return


    def __duplicate(self):
        """Create copy of files/directories"""
        if 'current' in self.request and 'target' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            target = self.__find(self.request['target'], curDir)
            if not curDir or not target:
                self.response['error'] = current.T('Invalid parameters')
                return
            if not self.__is_allowed(target, 'read') or \
               not self.__is_allowed(curDir, 'write'):
                self.response['error'] = current.T('Access denied')
            newName = self.__unique_name(target)
            if not self.__copy(target, newName):
                self.response['error'] = current.T('Unable to create file copy')
                return

        self.get_content(curDir, True)
        return


    def __resize(self):
        """Scale image size"""
        if not (
            'current' in self.request and 'target' in self.request
            and 'width' in self.request and 'height' in self.request
            ):
            self.response['error'] = current.T('Invalid parameters')
            return

        width = int(self.request['width'])
        height = int(self.request['height'])
        curDir = self.__find_dir(self.request['current'], None)
        curFile = self.__find(self.request['target'], curDir)

        if width < 1 or height < 1 or not curDir or not curFile:
            self.response['error'] = current.T('Invalid parameters')
            return
        if not self.__is_allowed(curFile, 'write'):
            self.response['error'] = current.T('Access denied')
            return
        if not self.get_mimetype(curFile).find('image') == 0:
            self.response['error'] = current.T('File is not an image')
            return

        self.debug('resize ' + curFile, str(width) + ':' + str(height))
        self.get_img_lib()
        try:
            im = self.im.open(curFile)
            imResized = im.resize((width, height), self.im.ANTIALIAS)
            imResized.save(curFile)
        except Exception, e:
            self.debug('resizeFailed_' + path, str(e))
            self.response['error'] = current.T('Unable to resize image')
            return

        self.response['select'] = [self.get_hash(curFile)]
        self.get_content(curDir, False)
        return


    def __thumbnails(self):
        """Create previews for images"""
        if 'current' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            if not curDir or curDir == self.options['tmbDir']:
                return False
        else:
            return False

        self.get_img_lib()
        if self.__can_create_tmb():
            if self.options['tmbAtOnce'] > 0:
                tmbMax = self.options['tmbAtOnce']
            else:
                tmbMax = 5
            self.response['current'] = self.get_hash(curDir)
            self.response['images'] = {}
            i = 0
            for f in os.listdir(curDir):
                path = os.path.join(curDir, f)
                fhash = self.get_hash(path)
                if self.__can_create_tmb(path) and \
                   self.__is_allowed(path, 'read'):
                    tmb = os.path.join(self.options['tmbDir'], fhash + '.png')
                    if not os.path.exists(tmb):
                        if self.__tmb(path, tmb):
                            self.response['images'].update({
                                fhash: self.__path2url(tmb)
                            })
                            i += 1
                if i >= tmbMax:
                    self.response['tmb'] = True
                    break
        else:
            return False

        return


    def __cwd(self, path):
        """Current Working Directory"""
        name = os.path.basename(path)
        if path == self.options['root']:
            name = self.options['rootAlias']
            root = True
        else:
            root = False

        if self.options['rootAlias']:
            basename = self.options['rootAlias']
        else:
            basename = os.path.basename(self.options['root'])

        rel = basename + path[len(self.options['root']):]

        self.response['cwd'] = {
            'hash': self.get_hash(path),
            'name': self.check_utf8(name),
            'mime': 'directory',
            'rel': self.check_utf8(rel),
            'size': 0,
            'date': datetime.fromtimestamp(os.stat(path).st_mtime).strftime("%d %b %Y %H:%M"),
            'read': True,
            'write': self.__is_allowed(path, 'write'),
            'rm': not root and self.__is_allowed(path, 'rm')
        }


    def __cdc(self, path):
        """Current Directory Content"""
        files = []
        dirs = []

        for f in sorted(os.listdir(path)):
            if not self.is_accepted(f): continue
            pf = os.path.join(path, f)
            info = {}
            info = self.__info(pf)
            info['hash'] = self.get_hash(pf)
            if info['mime'] == 'directory':
                dirs.append(info)
            else:
                files.append(info)

        dirs.extend(files)
        self.response['cdc'] = dirs


    def __info(self, path):
        mime = ''
        filetype = 'file'
        if os.path.isfile(path): filetype = 'file'
        if os.path.isdir(path): filetype = 'dir'
        if os.path.islink(path): filetype = 'link'

        stat = os.lstat(path)
        statDate = datetime.fromtimestamp(stat.st_mtime)

        fdate = ''
        if stat.st_mtime >= self.today:
            fdate = 'Today ' + statDate.strftime("%H:%M")
        elif stat.st_mtime >= self.yesterday and stat.st_mtime < self.today:
            fdate = 'Yesterday ' + statDate.strftime("%H:%M")
        else:
            fdate = statDate.strftime("%d %b %Y %H:%M")

        info = {
            'name': self.check_utf8(os.path.basename(path)),
            'hash': self.get_hash(path),
            'mime': 'directory' if filetype == 'dir' else self.get_mimetype(path),
            'date': fdate,
            'size': self.__dir_size(path) if filetype == 'dir' else stat.st_size,
            'read': self.__is_allowed(path, 'read'),
            'write': self.__is_allowed(path, 'write'),
            'rm': self.__is_allowed(path, 'rm')
        }

        if filetype == 'link':
            lpath = self.__readlink(path)
            if not lpath:
                info['mime'] = 'symlink-broken'
                return info

            if os.path.isdir(lpath):
                info['mime'] = 'directory'
            else:
                info['parent'] = self.get_hash(os.path.dirname(lpath))
                info['mime'] = self.get_mimetype(lpath)

            if self.options['rootAlias']:
                basename = self.options['rootAlias']
            else:
                basename = os.path.basename(self.options['root'])

            info['link'] = self.get_hash(lpath)
            info['linkTo'] = basename + lpath[len(self.options['root']):]
            info['read'] = info['read'] and self.__is_allowed(lpath, 'read')
            info['write'] = info['write'] and self.__is_allowed(lpath, 'write')
            info['rm'] = self.__is_allowed(lpath, 'rm')
        else:
            lpath = False

        if not info['mime'] == 'directory':
            if self.options['fileURL'] and info['read'] is True:
                if lpath:
                    info['url'] = self.__path2url(lpath)
                else:
                    info['url'] = self.__path2url(path)
            if info['mime'][0:5] == 'image':
                if self.__can_create_tmb():
                    dim = self.get_img_size(path)
                    if dim:
                        info['dim'] = dim
                        info['resize'] = True

                    # if we are in tmb dir, files are thumbs itself
                    if os.path.dirname(path) == self.options['tmbDir']:
                        info['tmb'] = self.__path2url(path)
                        return info

                    tmb = os.path.join(self.options['tmbDir'], 
                                       info['hash'] + '.png')

                    if os.path.exists(tmb):
                        tmbUrl = self.__path2url(tmb)
                        info['tmb'] = tmbUrl
                    else:
                        self.response['tmb'] = True

        return info


    def __tree(self, path):
        """Return directory tree starting from path"""

        if not os.path.isdir(path): return ''
        if os.path.islink(path): return ''

        if path == self.options['root'] and self.options['rootAlias']:
            name = self.options['rootAlias']
        else:
            name = os.path.basename(path)
        tree = {
            'hash': self.get_hash(path),
            'name': self.check_utf8(name),
            'read': self.__is_allowed(path, 'read'),
            'write': self.__is_allowed(path, 'write'),
            'dirs': []
        }

        if self.__is_allowed(path, 'read'):
            for d in sorted(os.listdir(path)):
                pd = os.path.join(path, d)
                if os.path.isdir(pd) and not os.path.islink(pd) and \
                self.is_accepted(d):
                    tree['dirs'].append(self.__tree(pd))

        return tree


    def __unique_name(self, path, copy = ' copy'):
        """Generate unique name for file copied file"""
        curDir = os.path.dirname(path)
        curName = os.path.basename(path)
        lastDot = curName.rfind('.')
        ext = newName = ''

        if not os.path.isdir(path) and \
           re.search(r'\..{3}\.(gz|bz|bz2)$', curName):
            pos = -7
            if curName[-1:] == '2':
                pos -= 1
            ext = curName[pos:]
            oldName = curName[0:pos]
            newName = oldName + copy
        elif os.path.isdir(path) or lastDot <= 0:
            oldName = curName
            newName = oldName + copy
            pass
        else:
            ext = curName[lastDot:]
            oldName = curName[0:lastDot]
            newName = oldName + copy

        pos = 0

        if oldName[-len(copy):] == copy:
            newName = oldName
        elif re.search(r'' + copy +'\s\d+$', oldName):
            pos = oldName.rfind(copy) + len(copy)
            newName = oldName[0:pos]
        else:
            newPath = os.path.join(curDir, newName + ext)
            if not os.path.exists(newPath):
                return newPath

        # if we are here then copy already exists or making copy of copy
        # we will make new indexed copy *black magic*
        idx = 1
        if pos > 0: idx = int(oldName[pos:])
        while True:
            idx += 1
            newNameExt = newName + ' ' + str(idx) + ext
            newPath = os.path.join(curDir, newNameExt)
            if not os.path.exists(newPath):
                return newPath
            # if idx >= 1000: break # possible loop

        return


    def __remove(self, target):
        """Internal remove procedure"""
        if not self.__is_allowed(target, 'rm'):
            self.set_error(target, 'Access denied')

        if not os.path.isdir(target):
            try:
                os.unlink(target)
                return True
            except:
                self.set_error(target, 'Remove failed')
                return False
        else:
            for i in os.listdir(target):
                if self.is_accepted(i):
                    self.__remove(os.path.join(target, i))

            try:
                os.rmdir(target)
                return True
            except:
                self.set_error(target, 'Remove failed')
                return False
        pass


    def __copy(self, src, dst):
        """Internal copy procedure"""
        dstDir = os.path.dirname(dst)
        if not self.__is_allowed(src, 'read'):
            self.set_error(src, 'Access denied')
            return False
        if not self.__is_allowed(dstDir, 'write'):
            self.set_error(dstDir, 'Access denied')
            return False
        if os.path.exists(dst):
            self.set_error(dst, 
                           'File or folder with the same name already exists')
            return False

        if not os.path.isdir(src):
            try:
                shutil.copyfile(src, dst)
                shutil.copymode(src, dst)
                return True
            except:
                self.set_error(src, 'Unable to copy files')
                return False
        else:
            try:
                os.mkdir(dst)
                shutil.copymode(src, dst)
            except:
                self.set_error(src, 'Unable to copy files')
                return False

            for i in os.listdir(src):
                newSrc = os.path.join(src, i)
                newDst = os.path.join(dst, i)
                if not self.__copy(newSrc, newDst):
                    self.set_error(newSrc, 'Unable to copy files')
                    return False

        return True


    def __find_dir(self, fhash, path):
        """Find directory by hash"""
        fhash = str(fhash)
        if not path:
            path = self.options['root']
            if fhash == self.get_hash(path):
                return path

        if not os.path.isdir(path):
            return None

        for d in os.listdir(path):
            pd = os.path.join(path, d)
            if os.path.isdir(pd) and not os.path.islink(pd):
                if fhash == self.get_hash(pd):
                    return pd
                else:
                    ret = self.__find_dir(fhash, pd)
                    if ret:
                        return ret

        return None


    def __find(self, fhash, parent):
        """Find file/dir by hash"""
        fhash = str(fhash)
        if os.path.isdir(parent):
            for i in os.listdir(parent):
                path = os.path.join(parent, i)
                if fhash == self.get_hash(path):
                    return path
        return None


    def __read(self):
        if 'current' in self.request and 'target' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            curFile = self.__find(self.request['target'], curDir)
            if curDir and curFile:
                if self.__is_allowed(curFile, 'read'):
                    self.response['content'] = open(curFile, 'r').read()
                else:
                    self.response['error'] = current.T('Access denied')
                return

        self.response['error'] = current.T('Invalid parameters')
        return


    def __edit(self):
        """Save content in file"""
        error = ''
        if 'current' in self.request and \
           'target' in self.request and \
           'content' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            curFile = self.__find(self.request['target'], curDir)
            error = curFile
            if curFile and curDir:
                if self.__is_allowed(curFile, 'write'):
                    try:
                        f = open(curFile, 'w+')
                        f.write(self.request['content'])
                        f.close()
                        self.response['target'] = self.__info(curFile)
                    except:
                        self.response['error'] = current.T('Unable to write to file')
                else:
                    self.response['error'] = current.T('Access denied')
            return

        self.response['error'] = current.T('Invalid parameters')
        return


    def __archive(self):
        """Compress files/directories to archive"""
        self.check_archivers()

        if (
            not self.options['archivers']['create'] 
            or not 'type' in self.request
            or not 'current' in self.request
            or not 'targets[]' in self.request
            or not 'name' in self.request
            ):
            self.response['error'] = current.T('Invalid parameters')
            return

        curDir = self.__find_dir(self.request['current'], None)
        archiveType = self.request['type']
        if (
            not archiveType in self.options['archivers']['create']
            or not archiveType in self.options['archiveMimes']
            or not curDir
            or not self.__is_allowed(curDir, 'write')
            ):
            self.response['error'] = current.T('Unable to create archive')
            return

        files = self.request['targets[]']
        if not isinstance(files, list):
            files = [files]

        realFiles = []
        for fhash in files:
            curFile = self.__find(fhash, curDir)
            if not curFile:
                self.response['error'] = current.T('File not found')
                return
            realFiles.append(os.path.basename(curFile))

        arc = self.options['archivers']['create'][archiveType]
        if len(realFiles) > 1:
            archiveName = self.request['name']
        else:
            archiveName = realFiles[0]
        archiveName += '.' + arc['ext']
        archiveName = self.__unique_name(archiveName, '')
        archivePath = os.path.join(curDir, archiveName)

        cmd = [arc['cmd']]
        for a in arc['argc'].split():
            cmd.append(a)
        cmd.append(archiveName)
        for f in realFiles:
            cmd.append(f)

        curCwd = os.getcwd()
        os.chdir(curDir)
        self.run_sub_process(cmd)
        os.chdir(curCwd)

        if os.path.exists(archivePath):
            self.get_content(curDir, False)
            self.response['select'] = [self.get_hash(archivePath)]
        else:
            self.response['error'] = current.T('Unable to create archive')

        return


    def __extract(self):
        """Uncompress archive"""
        if not 'current' in self.request or not 'target' in self.request:
            self.response['error'] = current.T('Invalid parameters')
            return

        curDir = self.__find_dir(self.request['current'], None)
        curFile = self.__find(self.request['target'], curDir)
        mime = self.get_mimetype(curFile)
        self.check_archivers()

        if (
            not mime in self.options['archivers']['extract']
            or not curDir
            or not curFile
            or not self.__is_allowed(curDir, 'write')
            ):
            self.response['error'] = current.T('Invalid parameters')
            return

        arc = self.options['archivers']['extract'][mime]

        cmd = [arc['cmd']]
        for a in arc['argc'].split():
            cmd.append(a)
        cmd.append(curFile)

        curCwd = os.getcwd()
        os.chdir(curDir)
        ret = self.run_sub_process(cmd)
        os.chdir(curCwd)

        if ret:
            self.get_content(curDir, True)
        else:
            self.response['error'] = current.T('Unable to extract files from archive')

        return


    def __tmb(self, path, tmb):
        """Internal thumbnail create procedure"""
        try:
            im = self.im.open(path).copy()
            size = self.options['tmbSize'], self.options['tmbSize']
            box = self.crop_tuple(im.size)
            if box:
                im = im.crop(box)
            im.thumbnail(size, self.im.ANTIALIAS)
            im.save(tmb, 'PNG')
        except Exception, e:
            self.debug('tmbFailed_' + path, str(e))
            return False
        return True


    def __rm_tmb(self, path):
        tmb = self.__tmb_path(path)
        if self.options['tmbDir']:
            if os.path.exists(tmb):
                try:
                    os.unlink(tmb)
                except:
                    pass




    def __readlink(self, path):
        """Read link and return real path if not broken"""
        target = os.readlink(path);
        if not target[0] == '/':
            target = os.path.join(os.path.dirname(path), target)
        target = os.path.normpath(target)
        if os.path.exists(target):
            if not target.find(self.options['root']) == -1:
                return target
        return False


    def __dir_size(self, path):
        total_size = 0
        if self.options['dirSize']:
            for dirpath, dirnames, filenames in os.walk(path):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    if os.path.exists(fp):
                        total_size += os.stat(fp).st_size
        else:
            total_size = os.lstat(path).st_size
        return total_size


    def __fbuffer(self, f, chunk_size = None):
        if chunk_size is None:
            chunk_size = self.options['uploadWriteChunk']
        while True:
            chunk = f.read(chunk_size)
            if not chunk: break
            yield chunk


    def __can_create_tmb(self, path = None):
        if self.options['imgLib'] and self.options['tmbDir']:
            if path is not None:
                mime = self.get_mimetype(path)
                if not mime[0:5] == 'image':
                    return False
            return True
        else:
            return False


    def __tmb_path(self, path):
        tmb = False
        if self.options['tmbDir']:
            if not os.path.dirname(path) == self.options['tmbDir']:
                tmb = os.path.join(self.options['tmbDir'], 
                                   self.get_hash(path) + '.png')
        return tmb


    def __path2url(self, path):
        curDir = path
        length = len(self.options['root'])
        url = str(self.options['URL'] + curDir[length:]).replace(os.sep, '/')

        try:
            import urllib
            url = urllib.quote(url, '/:~')
        except:
            pass
        return url


    def __is_allowed(self, path, access):
        if not os.path.exists(path):
            return False

        if access == 'read':
            if not os.access(path, os.R_OK):
                self.set_error(path, access)
                return False
        elif access == 'write':
            if not os.access(path, os.W_OK):
                self.set_error(path, access)
                return False
        elif access == 'rm':
            if not os.access(os.path.dirname(path), os.W_OK):
                self.set_error(path, access)
                return False
        else:
            return False

        path = path[len(os.path.normpath(self.options['root'])):]
        for ppath in self.options['perms']:
            regex = r'' + ppath
            if re.search(regex, path) and access in self.options['perms'][ppath]:
                return self.options['perms'][ppath][access]

        return self.options['defaults'][access]