# -*- coding: utf-8 -*-
#
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os
import os.path
import re
import shutil
import time
import urllib2
from datetime import datetime

import pysvn

from gluon import current
from gluon.storage import Storage

from .base import BaseDriver

SvnOptions =  {
        'root': '',
        'URL': '',
        'rootAlias': 'Home',
        'dotFiles': False,
        'dirSize': True,
        'fileMode': 0644,
        'dirMode': 0755,
        'imgLib': 'auto',
        'tmbDir': '.tmb',
        'tmbAtOnce': 5,
        'tmbSize': 48,
        'fileURL': True,
        'uploadMaxSize': 256,
        'uploadWriteChunk': 8192,
        'uploadAllow': [],
        'uploadDeny': [],
        'uploadOrder': ['deny', 'allow'],
        # 'aclObj': None, # TODO
        # 'aclRole': 'user', # TODO
        'defaults': {
            'read': True,
            'write': True,
            'rm': True
        },
        'perms': {},
        'archiveMimes': {},
        'archivers': {},
        'disabled': [],
        'debug': False,
        'repo_server': '',
        'repo_path': '',
        'repo_type': '',
        'repo_user': '', 
        'repo_pswd': ''
    }

class SvnDriver(BaseDriver):
    """Connector for elFinder"""
    
    def __init__(self, opts):
        options = SvnOptions
        for opt in opts:
            options[opt] = opts.get(opt)          
        BaseDriver.__init__(self, options)
        
        self.commands = {
            'open':    '__open',
            'reload': '__reload',
            'ping': '__ping'
        }

        self.options['URL'] = self.options['URL'].rstrip('/')
        self.debug('URL', self.options['URL'])
        self.debug('root', self.options['root'])
                
        # Svn connection
        self.svn_path = urllib2.urlparse.urljoin(self.options['repo_server'],
                                                 self.options['repo_path'])
        
        self.__set_connection()
        
            
    def __set_connection(self):          
        try:        
            self.client = pysvn.Client() 
            self.client.set_default_username(self.options['repo_user'])
            self.client.set_default_password(self.options['repo_pswd'])
            
#            def get_login( realm, username, may_save ):                
#                return True, self.options['repo_user'], self.options['repo_pswd'] , False
#            self.client.callback_get_login = get_login
            return True
        except Exception, e:
            self.response['error'] = current.T('Svn Connection Failed')
            self.debug('exception', str(e)) 
            return False
        

    def get_content(self, path, tree):
        """CWD + CDC + maybe(TREE)"""
        self.__cwd(path, tree)

        # Tree
        if tree:
            self.response['tree'] = self.__tree()       

    def _main_process(self, httpRequest):
        for field in self.httpAllowedParameters:
            if field in httpRequest:
                self.request[field] = httpRequest[field]
    
        if 'cmd' in self.request:
            if self.request['cmd'] in self.commands:
                cmd = self.commands[self.request['cmd']]
                func = getattr(self, '_' + self.__class__.__name__ + cmd, None)
                if callable(func):
                    try:
                        func()
                    except Exception, e:
                        if not self.response['error']:
                            self.response['error'] = current.T('Command Failed')
                        self.debug('exception', str(e))
            else:
                self.response['error'] = current.T('Unknown command')
        else:
            self.__open()

        if 'init' in self.request:
            self.check_archivers()
            self.response['disabled'] = self.options['disabled']
            if not self.options['fileURL']:
                url = ''
            else:
                url = self.options['URL']
            self.response['params'] = {
                'dotFiles': self.options['dotFiles'],
                'uplMaxSize': str(self.options['uploadMaxSize']) + 'M',
                'archives': self.options['archiveMimes'],
                'extract': self.options['archivers']['extract'].keys(),
                'url': url
            }
                
    def __get_directory_format(self, path = ''):
        r_path = path.replace(self.svn_path + '/', '')
        content = {'cwd': Storage(path=r_path,
                                  type='dir',
                                  size=0,
                                  last_modif=0), 
                   'collection': []}
        if self.client:
            try:
                drv_content = self.__ls(path)
                for item in drv_content:
                    # Relative path
                    r_path = item.name.replace(self.svn_path + '/', '')
                    tmp = Storage(path=r_path,
                                  type=item.kind,
                                  size=item.size or 0,
                                  last_modif=item.time)
                    content['collection'].append(tmp)
            except Exception, e:
                pass
        return content
        
    def __ls(self, path=None):
        if path is None:
            path = self.svn_path
        else:
            path = '%s/%s' % (self.svn_path, path)
        try:
            data = self.client.ls(path)
        except Exception, e:
            self.response['error'] = current.T('Authorization failed')
            data = None
        
        return data
    
    def __open(self):
        """Open file or directory"""
        # try to open file
        path = ''

        if 'target' in self.request:
            if self.request['target'] != path:
                target = self.__find_dir(self.request['target'], None)
                if target is None:
                    self.response['error'] = current.T('Invalid parameters')
                else:
                    path = target

        self.get_content(path, 'tree' in self.request)



    def __cwd(self, path, tree):
        """Current Working Directory"""
        # Get data
        cwd_data = self.__get_directory_format(path)        
        
        #
        # CWD data
        elem = cwd_data['cwd']
        if elem.path == '':
            name = self.options['rootAlias']
            root = True
        else:
            name = ''
            root = False

        if self.options['rootAlias']:
            basename = self.options['rootAlias']
        else:
            basename = os.path.basename(self.options['repo_path'])

        rel = basename + path[len(self.options['repo_path']):]

        # Datetime
        try:
            date = datetime.fromtimestamp(elem.last_modif)
        except Exception, e:
            date = self.today
            
        self.response['cwd'] = {
            'hash': self.get_hash(elem.path),
            'name': self.check_utf8(name),
            'mime': 'directory',
            'rel': self.check_utf8(rel),
            'size': elem.size,
            'date': date.strftime("%d %b %Y %H:%M"),
            'read': True,
            'write': self.__is_allowed(path, 'write'),
            'rm': not root and self.__is_allowed(path, 'rm')
        }
        
        #
        # Current Directory Content        
        collection = cwd_data['collection']
        files = []
        dirs = []          
           
        for item in collection:
            info = {}
            info = self.__info(item)
            if info['mime'] == 'directory':
                dirs.append(info)
            else:
                files.append(info)

        dirs.extend(files)

        self.response['cdc'] = dirs

    def __info(self, item):
        mime = ''
        filetype = 'file'
        if item.type == pysvn.node_kind.dir: filetype = 'dir'

        fdate = ''
        try:            
            stat = item.last_modif
            statDate = datetime.fromtimestamp(item.last_modif)
            
            if stat >= self.today:
                fdate = 'Today ' + statDate.strftime("%H:%M")
            elif stat >= self.yesterday and stat < self.today:
                fdate = 'Yesterday ' + statDate.strftime("%H:%M")
            else:
                fdate = statDate.strftime("%d %b %Y %H:%M")
        except:
            pass
        
        c_path = item.path
        c_hash = self.get_hash(item.path)
        info = {
            'name': self.check_utf8(os.path.basename(item.path)),
            'hash': c_hash,
            'mime': 'directory' if filetype == 'dir' else self.get_mimetype(item.path),
            'date': fdate,
            'size': item.size,
            'read': self.__is_allowed(item.path, 'read'),
            'write': self.__is_allowed(item.path, 'write'),
            'rm': self.__is_allowed(item.path, 'rm')
        }

        if not info['mime'] == 'directory':
            if self.options['fileURL'] and info['read'] is True:
                info['url'] = self.__path2url(item.path)
        return info


    def __tree(self, path=''):
        """Return directory tree starting from path""" 
        cwd_data = self.__get_directory_format(path)        
        
        #
        # CWD data
        elem = cwd_data['cwd']    
        collection = cwd_data['collection']
        
        #
        # Tree processing
        if elem.path == '':
            name = self.options['rootAlias']
        else:
            name = ''

        tree = {
            'hash': self.get_hash(elem.path),
            'name': self.check_utf8(name),
            'read': self.__is_allowed(elem.path, 'read'),
            'write': self.__is_allowed(elem.path, 'write'),
            'dirs': []
        }
        
        self.tree_level = 3
        for item in collection:
            if item.type == pysvn.node_kind.dir:
                path = urllib2.urlparse.urljoin(elem.path, item.path)
                tree['dirs'].append(self.__svn__tree(path))

        return tree

    def __svn__tree(self, path):        
        cwd_data = self.__get_directory_format(path)  
        #
        # CWD data
        if cwd_data['cwd'] is not None:
            elem = cwd_data['cwd']
            if elem.path == '':
                name = self.options['rootAlias']
            else:
                name = elem.path.split('/')[-1]
    
            tree = {
                'hash': self.get_hash(elem.path),
                'name': self.check_utf8(name),
                'read': self.__is_allowed(elem.path, 'read'),
                'write': self.__is_allowed(elem.path, 'write'),
                'dirs': []
            }
            
            # Recursive directory search
            collection = cwd_data['collection'] 
            for item in collection:
                if item.type == pysvn.node_kind.dir:
                    tree['dirs'].append(self.__svn__tree(item.path))
            
            return tree
        else:
            return '---'
        

    def __unique_name(self, path, copy = ' copy'):
        """Generate unique name for file copied file"""
        curDir = os.path.dirname(path)
        curName = os.path.basename(path)
        lastDot = curName.rfind('.')
        ext = newName = ''

        if not os.path.isdir(path) and re.search(r'\..{3}\.(gz|bz|bz2)$', curName):
            pos = -7
            if curName[-1:] == '2':
                pos -= 1
            ext = curName[pos:]
            oldName = curName[0:pos]
            newName = oldName + copy
        elif os.path.isdir(path) or lastDot <= 0:
            oldName = curName
            newName = oldName + copy
            pass
        else:
            ext = curName[lastDot:]
            oldName = curName[0:lastDot]
            newName = oldName + copy

        pos = 0

        if oldName[-len(copy):] == copy:
            newName = oldName
        elif re.search(r'' + copy +'\s\d+$', oldName):
            pos = oldName.rfind(copy) + len(copy)
            newName = oldName[0:pos]
        else:
            newPath = os.path.join(curDir, newName + ext)
            if not os.path.exists(newPath):
                return newPath

        # if we are here then copy already exists or making copy of copy
        # we will make new indexed copy *black magic*
        idx = 1
        if pos > 0: idx = int(oldName[pos:])
        while True:
            idx += 1
            newNameExt = newName + ' ' + str(idx) + ext
            newPath = os.path.join(curDir, newNameExt)
            if not os.path.exists(newPath):
                return newPath
            # if idx >= 1000: break # possible loop

        return



    def __find_dir(self, fhash, path):
        """Find directory by hash"""
        fhash = str(fhash)
        if not path:
            path = ''
            n_hash = self.get_hash(path)
            if fhash == n_hash:
                return path 
                    
        return self.__svn__find_dir(fhash, path)

    def __svn__find_dir(self, fhash, path):
        t_path = None
        collection = self.__get_directory_format(path)['collection']              
        # Recursive directory search
        for item in collection:
            if item.type == pysvn.node_kind.dir:                    
                n_hash = self.get_hash(item.path)
                if fhash == n_hash:
                    return item.path
                # It is not curret path
                t_path = self.__svn__find_dir(fhash, item.path) 
                if t_path is not None:
                    return t_path
            
        return t_path


    def __find(self, fhash, parent):
        """Find file/dir by hash"""
        t_path, type = self.__svn__find_in_dir(fhash, parent)
        
        if type == 'directory':
            t_path = t_path + '/'
            
        return t_path
    
    def __svn__find_in_dir(self, fhash, path):
        t_path = None
        type = None
        collection = self.__get_directory_format(path)['collection']              
        # Recursive directory search
        for item in collection:   
            type = 'file'      
            if item.type == pysvn.node_kind.dir:  
                type = 'directory'  
                            
            n_hash = self.get_hash(item.path)
            if fhash == n_hash:
                return item.path, type
            
            # It is not curret path
            t_path = self.__svn__find_dir(fhash, item.path) 
            if t_path is not None:
                return t_path, type 
            
        return t_path, type


    def __read(self):
        if 'current' in self.request and 'target' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            curFile = self.__find(self.request['target'], curDir)
            if curDir and curFile:
                if self.__is_allowed(curFile, 'read'):
                    self.response['content'] = open(curFile, 'r').read()
                else:
                    self.response['error'] = current.T('Access denied')
                return

        self.response['error'] = current.T('Invalid parameters')
        return

    def __dir_size(self, path):
        total_size = 0
        if self.options['dirSize']:
            for dirpath, dirnames, filenames in os.walk(path):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    if os.path.exists(fp):
                        total_size += os.stat(fp).st_size
        else:
            total_size = os.lstat(path).st_size
        return total_size


    def __fbuffer(self, f, chunk_size = None):
        if chunk_size is None:
            chunk_size = self.options['uploadWriteChunk']
        while True:
            chunk = f.read(chunk_size)
            if not chunk: break
            yield chunk

    def __path2url(self, path):
        curDir = path
        length = len(self.options['root'])
        url = str(self.options['URL'] + '/' + curDir[length:]).replace(os.sep, '/')

        try:
            import urllib
            url = urllib.quote(url, '/:~')
        except:
            pass
        return url


    def __is_allowed(self, path, access):
        # In svn connection, allways true
        return True