# -*- coding: utf-8 -*-
#
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os
import os.path
import re
import shutil
import time
import urllib2
from datetime import datetime

from gluon import current
from gluon.storage import Storage

import python_webdav.connection as conn

from .base import BaseDriver

WebdavOptions =  {
        'root': '',
        'URL': '',
        'rootAlias': 'Home',
        'dotFiles': False,
        'dirSize': True,
        'fileMode': 0644,
        'dirMode': 0755,
        'imgLib': 'auto',
        'tmbDir': '.tmb',
        'tmbAtOnce': 5,
        'tmbSize': 48,
        'fileURL': True,
        'uploadMaxSize': 256,
        'uploadWriteChunk': 8192,
        'uploadAllow': [],
        'uploadDeny': [],
        'uploadOrder': ['deny', 'allow'],
        # 'aclObj': None, # TODO
        # 'aclRole': 'user', # TODO
        'defaults': {
            'read': True,
            'write': True,
            'rm': True
        },
        'perms': {},
        'archiveMimes': {},
        'archivers': {},
        'disabled': [],
        'debug': False,
        'dav': {'path': '.', 
                'port': '80',
                'realm': ''},
        'dav_server': '', 
        'dav_path': '', 
        'dav_user': '', 
        'dav_pswd': ''
    }

class WebdavDriver(BaseDriver):
    """Connector for elFinder"""
    
    def __init__(self, opts):
        options = WebdavOptions
        for opt in opts:
            options[opt] = opts.get(opt)          
        BaseDriver.__init__(self, options)
        
        self.commands = {
            'open':    '__open',
            'reload': '__reload',
            'mkdir': '__mkdir',
            'mkfile': '__mkfile',
            'upload': '__upload',
            'rm': '__rm',
#            'extract': '__extract',
#            'archive': '__archive',
            'ping': '__ping'
        }

        self.options['URL'] = self.options['URL'].rstrip('/')
        self.debug('URL', self.options['URL'])
        self.debug('root', self.options['root'])
        
        # Checks dav path
        if current._config.get('webdav_type') != 'pywebdav':
            if self.options['dav_server'][-1] != '/':
                self.options['dav_server'] += '/'
                
        # Webdav connection
        self.dav_base = None
        self.dav_root = urllib2.urlparse.urljoin(self.options['dav_server'],
                                            self.options['dav_path'])
        self._connection_settings = dict(host=self.dav_root,
                                         path=self.options['dav_path'],
                                         port=self.options['dav']['port'], 
                                         realm=self.options['dav']['realm'],
                                         username=self.options['dav_user'],
                                         password=self.options['dav_pswd'],)
        
            
    def __set_connection(self, path=None):          
        try:
            if path is not None:
                self._connection_settings['path'] = path
            self.connection = conn.Connection(self._connection_settings)
            self.client = conn.Client()
            return True
        except Exception, e:
            self.response['error'] = current.T('Webdav Connection Failed')
            self.debug('exception', str(e)) 
            return False
        

    def get_content(self, path, tree):
        """CWD + CDC + maybe(TREE)"""
        self.__cwd(path, tree)

        # Tree
        if tree:
            self.response['tree'] = self.__tree()       

    def _main_process(self, httpRequest):
        for field in self.httpAllowedParameters:
            if field in httpRequest:
                self.request[field] = httpRequest[field]
    
        if 'cmd' in self.request:
            if self.request['cmd'] in self.commands:
                cmd = self.commands[self.request['cmd']]
                func = getattr(self, '_' + self.__class__.__name__ + cmd, None)
                if callable(func):
                    try:
                        func()
                    except Exception, e:
                        if not self.response['error']:
                            self.response['error'] = current.T('Command Failed')
                        self.debug('exception', str(e))
            else:
                self.response['error'] = current.T('Unknown command')
        else:
            self.__open()

        if 'init' in self.request:
            self.check_archivers()
            self.response['disabled'] = self.options['disabled']
            if not self.options['fileURL']:
                url = ''
            else:
                url = self.options['URL']
            self.response['params'] = {
                'dotFiles': self.options['dotFiles'],
                'uplMaxSize': str(self.options['uploadMaxSize']) + 'M',
                'archives': self.options['archiveMimes'],
                'extract': self.options['archivers']['extract'].keys(),
                'url': url
            }
                
    def __get_directory_format(self, path = ''):
        content = {'cwd': None, 'collection': []}
        if self.__set_connection(path):
            try:
                wdav_content = self.__ls(path)
                for item in wdav_content:
                    if current._config.get('webdav_type') == 'pywebdav':
                        item_path = item[0].rstrip('/')
                    else:
                        item_path = item[0].rstrip('/')
                        if self.dav_base is None:
                            self.dav_base = self.dav_root.replace(item_path, '')
                        item_path = "%s%s" % (self.dav_base, item_path)
                    # Relative path
                    r_path = item_path.replace(self.dav_root, '')
                    tmp = Storage(path=r_path.lstrip('/'),
                                  type=item[1],
                                  size=item[2] or 0,
                                  last_modif=item[3])
                    if item_path == path:
                        content['cwd'] = tmp
                    else:
                        content['collection'].append(tmp)
            except Exception, e:
                self.response['error'] = current.T('Webdav Connection Failed')
                self.debug('exception', str(e))
        return content
        
    def __ls(self, path=''):
        """
            :param path: Path of the directory to list
            :type path: String

            :param list_format: This is the format for the directory listing

            The format symbols are:
                * T - Type (resourcetype)
                * D - Date Time (creationdate)
                * F - Filename (href)
                * M - Last modified time (getlastmodified)
                * A - Attributes (executable)
                * E - ETag (getetag)
                * C - Content type (getcontenttype)

            :type list_format: List

            :param separator: Separator to use for formatting output
            :type separator: String

            :param display: Whether or not to print the output
            :type display: Boolean

        """
        # Format Map
        format_map = {'T': 'resourcetype',
                      'D': 'creationdate',
                      'F': 'href',
                      'M': 'getlastmodified',
                      'A': 'executable',
                      'E': 'getetag',
                      'C': 'getcontenttype',
#                      'L': 'locks',
                      'S': 'getcontentlength'}
        
        list_format=('F','C','S','M','T','A','E')
        # Get the properties for the given path
        if not path:
            path  = self.connection.path
        props = self.client.get_properties(self.connection, path)
        property_lists = []
        for prop in props:
            format_string = ''
            formatted_list = [getattr(prop, format_map[symbol]) for symbol in list_format]
            property_lists.append(formatted_list)
            
        return property_lists
    
    def __open(self):
        """Open file or directory"""
        # try to open file
        path = ''

        if 'target' in self.request:
            if self.request['target'] != path:
                target = self.__find_dir(self.request['target'], None)
                if target is None:
                    self.response['error'] = current.T('Invalid parameters')
                else:
                    path = target

        self.get_content(path, 'tree' in self.request)


    def __rename(self):
        """Rename file or dir"""
        self.get_content('', True)
        # Copy with new name
        # Remove older directory


    def __mkdir(self):
        """Create new directory"""
        current = None
        path = None
        newDir = None
        if 'name' in self.request and 'current' in self.request:
            name = self.request['name']
            current = self.request['current']
            cur_path = self.__find_dir(current, None)
            if cur_path == '':
                path = self.dav_root
            else:            
                path = "%s/%s" % (self.dav_root, cur_path)            
            newDir = os.path.join(path, name)

        if not path:
            self.response['error'] = current.T('Invalid parameters')
        elif not self.__is_allowed(path, 'write'):
            self.response['error'] = current.T('Access denied')
        elif not self.check_name(name):
            self.response['error'] = current.T('Invalid name')
#        elif os.path.exists(newDir):
#            self.response['error'] = 'File or folder with the same name already exists'
        else:
            try:
                self.__dav_mkdir(newDir)
                self.response['select'] = [self.get_hash(name)]
                self.get_content(cur_path, True)
            except Exception, e:
                self.response['error'] = current.T('Unable to create folder')

    def __dav_mkdir(self, new_path):
        if self.__set_connection():
            try:
                response,content = self.connection.send_mkcol(new_path)
                if not response['status'] in ['200', '201']:
                    if response['status'] == '401':
                        self.response['error'] = current.T('Authentication required')
            except Exception, e:
                raise e

    def __rm(self):
        """Delete files and directories"""
        current = rmList = None
        curDir = rmFile = None
        if 'current' in self.request and 'targets[]' in self.request:
            current = self.request['current']
            rmList = self.request['targets[]']
            curDir = self.__find_dir(current, None)
            if curDir == '':
                cur_path = self.dav_root
            else:            
                cur_path = "%s/%s" % (self.dav_root, curDir)  

        if not rmList or not cur_path:
            self.response['error'] = current.T('Invalid parameters')
            return False

        if not isinstance(rmList, list):
            rmList = [rmList]

        for rm in rmList:
            rmFile = self.__find(rm, cur_path)
            if rmFile == '':
                rm_file = self.dav_root
            else:            
                rm_file = "%s/%s" % (self.dav_root, rmFile)  
            
            if not rm_file: continue
            self.__dav_rm(rm_file)
            
        self.get_content(curDir, True)

    def __dav_rm(self, rm_path):
        if self.__set_connection():
            try:
                response,content = self.connection.send_delete(rm_path)
                if response['status'] != '200':
                    if response['status'] == '401':
                        self.response['error'] = current.T('Authentication required')
            except Exception, e:
                raise e


    def __upload(self):
        """Upload files"""
        try: # Windows needs stdio set for binary mode.
            import msvcrt
            msvcrt.setmode (0, os.O_BINARY) # stdin  = 0
            msvcrt.setmode (1, os.O_BINARY) # stdout = 1
        except ImportError:
            pass
        import StringIO
        
        if 'current' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            if curDir == '':
                cur_path = self.dav_root
            else:            
                cur_path = "%s/%s" % (self.dav_root, curDir)  
                
            if not cur_path:
                self.response['error'] = current.T('Invalid parameters')
                return
            if not self.__is_allowed(curDir, 'write'):
                self.response['error'] = current.T('Access denied')
                return
            if not 'upload[]' in self.request:
                self.response['error'] = current.T('No file to upload')
                return

            upFiles = self.request['upload[]']
            # invalid format
            # must be dict('filename1': 'filedescriptor1', 'filename2': 'filedescriptor2', ...)
            if not isinstance(upFiles, dict):
                self.response['error'] = current.T('Invalid parameters')
                return

            self.response['select'] = []
            total = 0
            upSize = 0
            maxSize = self.options['uploadMaxSize'] * 1024 * 1024
            for name, data in upFiles.iteritems():
                if name:
                    total += 1
                    name = os.path.basename(name)
                    if not self.check_name(name):
                        self.set_error(name, 'Invalid name')
                    else:
                        name = os.path.join(curDir, name)
                        try:
                            f = StringIO.StringIO()
                            for chunk in self.__fbuffer(data):
                                f.write(chunk)
                            file_contents = f.getvalue()    
                            f.close()
                            f_path = "%s/%s" % (self.dav_root, name) 
                            self.__dav_mkfile(f_path, file_contents)
                        except:
                            self.set_error(name, 'Unable to save uploaded file')
#                        if upSize > maxSize:
#                            try:
#                                os.unlink(name)
#                                self.set_error(name, 'File exceeds the maximum allowed filesize')
#                            except:
#                                pass
#                                # TODO ?
#                                self.set_error(name, 'File was only partially uploaded')
#                            break

            if self.errorData:
                if len(self.errorData) == total:
                    self.response['error'] = current.T('Unable to upload files')
                else:
                    self.response['error'] = current.T('Some files was not uploaded')

            self.get_content(curDir, False)
            return
        
    def __dav_mkfile(self, new_path, content):
        if self.__set_connection():
            try:
                response,content = self.connection.send_put(new_path, content)
                if not response['status'] in ['200', '201']:
                    if response['status'] == '401':
                        self.response['error'] = current.T('Authentication required')
                    else:
                        if int(self.response['error']) > 400:
                            self.response['error'] = current.T('Dav server error')
            except Exception, e:
                self.response['error'] = current.T('Dav server error')
                raise e


    def __cwd(self, path, tree):
        """Current Working Directory"""
        # Get data
        if path == '':
            path = self.dav_root
        else:            
            path = "%s/%s" % (self.dav_root, path)
            
        cwd_data = self.__get_directory_format(path)        
        
        #
        # CWD data
        elem = cwd_data['cwd']
        if elem.path == '':
            name = self.options['rootAlias']
            root = True
        else:
            name = ''
            root = False

        if self.options['rootAlias']:
            basename = self.options['rootAlias']
        else:
            basename = os.path.basename(self.options['dav_path'])

        rel = basename + path[len(self.options['dav_path']):]

        # Datetime
        try:
            date = datetime.strptime(elem.last_modif, 
                                     "%a, %d %b %Y %H:%M:%S GMT")
        except Exception, e:
            date = self.today
            
        self.response['cwd'] = {
            'hash': self.get_hash(elem.path),
            'name': self.check_utf8(name),
            'mime': 'directory',
            'rel': self.check_utf8(rel),
            'size': elem.size,
            'date': date.strftime("%d %b %Y %H:%M"),
            'read': True,
            'write': self.__is_allowed(path, 'write'),
            'rm': not root and self.__is_allowed(path, 'rm')
        }
        
        #
        # Current Directory Content        
        collection = cwd_data['collection']
        files = []
        dirs = []          
           
        for item in collection:
            info = {}
            info = self.__info(item)
            if info['mime'] == 'directory':
                dirs.append(info)
            else:
                files.append(info)

        dirs.extend(files)

        self.response['cdc'] = dirs

    def __info(self, item):
        mime = ''
        filetype = 'file'
        if item.type == 'httpd/unix-directory': filetype = 'dir'

        fdate = ''
        try:
            time_tuple = time.strptime(item.last_modif, 
                                       "%a, %d %b %Y %H:%M:%S GMT")
            stat = time.mktime(time_tuple)
            statDate = datetime.strptime(item.last_modif, 
                                         "%a, %d %b %Y %H:%M:%S GMT")
            if stat >= self.today:
                fdate = 'Today ' + statDate.strftime("%H:%M")
            elif stat >= self.yesterday and stat < self.today:
                fdate = 'Yesterday ' + statDate.strftime("%H:%M")
            else:
                fdate = statDate.strftime("%d %b %Y %H:%M")
        except:
            pass
        
        c_path = item.path
        c_hash = self.get_hash(item.path)
        info = {
            'name': self.check_utf8(os.path.basename(item.path)),
            'hash': c_hash,
            'mime': 'directory' if filetype == 'dir' else self.get_mimetype(item.path),
            'date': fdate,
            'size': item.size,
            'read': self.__is_allowed(item.path, 'read'),
            'write': self.__is_allowed(item.path, 'write'),
            'rm': self.__is_allowed(item.path, 'rm')
        }

        if not info['mime'] == 'directory':
            if self.options['fileURL'] and info['read'] is True:
                info['url'] = self.__path2url(item.path)
        return info


    def __tree(self, path=''):
        # Get data
        if path == '':
            path = self.dav_root
        else:            
            path = "%s/%s" % (self.dav_root, path)
        """Return directory tree starting from path""" 
        cwd_data = self.__get_directory_format(path)        
        
        #
        # CWD data
        elem = cwd_data['cwd']    
        collection = cwd_data['collection']
        
        #
        # Tree processing
        if elem.path == '':
            name = self.options['rootAlias']
        else:
            name = ''

        tree = {
            'hash': self.get_hash(elem.path),
            'name': self.check_utf8(name),
            'read': self.__is_allowed(elem.path, 'read'),
            'write': self.__is_allowed(elem.path, 'write'),
            'dirs': []
        }
        
        for item in collection:
            if item.type == 'httpd/unix-directory':
                host = self.dav_root.rstrip('/')
                if elem.path == '':
                    path = "%s/%s" % (host, item.path)
                else:
                    path = "%s/%s/%s" % (host, elem.path, item.path)
                tree['dirs'].append(self.__dav__tree(path))

        return tree

    def __dav__tree(self, path):
        if path == '':
            path = self.dav_root
        else:            
            if not path.startswith(self.dav_root):
                path = '%s/%s' % (self.dav_root, path)  
        cwd_data = self.__get_directory_format(path)  
        #
        # CWD data
        if cwd_data['cwd'] is not None:
            elem = cwd_data['cwd']
            if elem.path == '':
                name = self.options['rootAlias']
            else:
                name = elem.path.split('/')[-1]
    
            tree = {
                'hash': self.get_hash(elem.path),
                'name': self.check_utf8(name),
                'read': self.__is_allowed(elem.path, 'read'),
                'write': self.__is_allowed(elem.path, 'write'),
                'dirs': []
            }
            
            # Recursive directory search
            collection = cwd_data['collection'] 
            for item in collection:
                if item.type == 'httpd/unix-directory':
#                    path = urllib2.urlparse.urljoin(elem.path, item.path)
                    tree['dirs'].append(self.__dav__tree(item.path))
            
            return tree
        else:
            return '---'
        

    def __unique_name(self, path, copy = ' copy'):
        """Generate unique name for file copied file"""
        curDir = os.path.dirname(path)
        curName = os.path.basename(path)
        lastDot = curName.rfind('.')
        ext = newName = ''

        if not os.path.isdir(path) and re.search(r'\..{3}\.(gz|bz|bz2)$', curName):
            pos = -7
            if curName[-1:] == '2':
                pos -= 1
            ext = curName[pos:]
            oldName = curName[0:pos]
            newName = oldName + copy
        elif os.path.isdir(path) or lastDot <= 0:
            oldName = curName
            newName = oldName + copy
            pass
        else:
            ext = curName[lastDot:]
            oldName = curName[0:lastDot]
            newName = oldName + copy

        pos = 0

        if oldName[-len(copy):] == copy:
            newName = oldName
        elif re.search(r'' + copy +'\s\d+$', oldName):
            pos = oldName.rfind(copy) + len(copy)
            newName = oldName[0:pos]
        else:
            newPath = os.path.join(curDir, newName + ext)
            if not os.path.exists(newPath):
                return newPath

        # if we are here then copy already exists or making copy of copy
        # we will make new indexed copy *black magic*
        idx = 1
        if pos > 0: idx = int(oldName[pos:])
        while True:
            idx += 1
            newNameExt = newName + ' ' + str(idx) + ext
            newPath = os.path.join(curDir, newNameExt)
            if not os.path.exists(newPath):
                return newPath
            # if idx >= 1000: break # possible loop

        return


    def __remove(self, target):
        """Internal remove procedure"""
        if not self.__is_allowed(target, 'rm'):
            self.set_error(target, 'Access denied')

        if not os.path.isdir(target):
            try:
                os.unlink(target)
                return True
            except:
                self.set_error(target, 'Remove failed')
                return False
        else:
            for i in os.listdir(target):
                if self.is_accepted(i):
                    self.__remove(os.path.join(target, i))

            try:
                os.rmdir(target)
                return True
            except:
                self.set_error(target, 'Remove failed')
                return False
        pass


    def __copy(self, src, dst):
        """Internal copy procedure"""
        dstDir = os.path.dirname(dst)
        if not self.__is_allowed(src, 'read'):
            self.set_error(src, 'Access denied')
            return False
        if not self.__is_allowed(dstDir, 'write'):
            self.set_error(dstDir, 'Access denied')
            return False
        if os.path.exists(dst):
            self.set_error(dst, 'File or folder with the same name already exists')
            return False

        if not os.path.isdir(src):
            try:
                shutil.copyfile(src, dst)
                shutil.copymode(src, dst)
                return True
            except:
                self.set_error(src, 'Unable to copy files')
                return False
        else:
            try:
                os.mkdir(dst)
                shutil.copymode(src, dst)
            except:
                self.set_error(src, 'Unable to copy files')
                return False

            for i in os.listdir(src):
                newSrc = os.path.join(src, i)
                newDst = os.path.join(dst, i)
                if not self.__copy(newSrc, newDst):
                    self.set_error(newSrc, 'Unable to copy files')
                    return False

        return True


    def __find_dir(self, fhash, path):
        """Find directory by hash"""
        fhash = str(fhash)
        if not path:
            path = '' #self.options['dav_server']
            n_hash = self.get_hash(path)
            if fhash == n_hash:
                return path 
                    
        return self.__dav__find_dir(fhash, path)

    def __dav__find_dir(self, fhash, path):
        if path == '':
            path = self.dav_root
        else:            
            path = "%s/%s" % (self.dav_root, path)
        t_path = None
        collection = self.__get_directory_format(path)['collection']              
        # Recursive directory search
        for item in collection:
            if item.type == 'httpd/unix-directory':                    
                n_hash = self.get_hash(item.path)
                if fhash == n_hash:
                    return item.path
                # It is not curret path
                t_path = self.__dav__find_dir(fhash, item.path) 
                if t_path is not None:
                    return t_path
            
        return t_path


    def __find(self, fhash, parent):
        """Find file/dir by hash"""
        t_path, type = self.__dav__find_in_dir(fhash, parent)
        
        if type == 'directory':
            t_path = t_path + '/'
            
        return t_path
    
    def __dav__find_in_dir(self, fhash, path):
        t_path = None
        type = None
        collection = self.__get_directory_format(path)['collection']              
        # Recursive directory search
        for item in collection:   
            type = 'file'      
            if item.type == 'httpd/unix-directory':  
                type = 'directory'  
                            
            n_hash = self.get_hash(item.path)
            if fhash == n_hash:
                return item.path, type
            
            # It is not curret path
            t_path = self.__dav__find_dir(fhash, item.path) 
            if t_path is not None:
                return t_path, type 
            
        return t_path, type


    def __read(self):
        if 'current' in self.request and 'target' in self.request:
            curDir = self.__find_dir(self.request['current'], None)
            curFile = self.__find(self.request['target'], curDir)
            if curDir and curFile:
                if self.__is_allowed(curFile, 'read'):
                    self.response['content'] = open(curFile, 'r').read()
                else:
                    self.response['error'] = 'Access denied'
                return

        self.response['error'] = 'Invalid parameters'
        return

    def __dir_size(self, path):
        total_size = 0
        return total_size

    def __fbuffer(self, f, chunk_size = None):
        if chunk_size is None:
            chunk_size = self.options['uploadWriteChunk']
        while True:
            chunk = f.read(chunk_size)
            if not chunk: break
            yield chunk

    def __path2url(self, path):
        curDir = path
        length = len(self.options['root'])
        url = str(self.options['URL'] + '/' + curDir[length:]).replace(os.sep, '/')

        try:
            import urllib
            url = urllib.quote(url, '/:~')
        except:
            pass
        return url


    def __is_allowed(self, path, access):
        # In webdav connection, allways true
        return True