# -*- coding: utf-8 -*- 

#########################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
# 
# Use epydoc format to document
# http://epydoc.sourceforge.net/manual-epytext.html
#########################################################################

import os
import shutil
import time

from mil.exceptions import FileNotExists
        
class File:
    perms = 0775
    
    """
    This class implements the basic statics methods to handle a file.
    """
    @staticmethod        
    def exists(filename):
        """
        Checks if a file exists.

        @param filename: Path to file.
        @return: If file exists returns True, else returns False.
        """
        try:
            return os.path.exists(filename)
        except IOError, (errno, errstr):
            #print("I/O Error (%s): %s <%s>" % (errno, errstr, filename))
            return False
        
    @staticmethod   
    def check_dir(folder):
        """ 
        Checks if a folder exists.

        @param dir: Path to folder.
        @return: If folder exists returns True, else returns False.
        """
        return os.path.isdir(folder)
            
    @staticmethod        
    def add(filename, data):
        """ 
        Write a content data into a file.
        
        It is a wrapper to File.save() function.

        @param filename: Path to file.
        @param data: Data to be included into file.
        @return: The same values as File.save().
        """
        if File.exists(filename) is False:
            File.touch(filename)
        File.save(filename, "%s"%data, 'a')
            
    @staticmethod        
    def write(filename, data):
        """ 
        Write a content data into a file. If file do not exists, it is created.
        
        It is a wrapper to File.save() function.

        @param filename: Path to file.
        @param data: Data to be included into file.
        @return: The same values as File.save().
        """
        if File.exists(filename) is False:
            File.touch(filename)
        return File.save(filename, "%s"%data)
            
    @staticmethod        
    def save(filename, data, type='w'):
        """
        Saves data into a file.

        @param filename: Path to file.
        @param data: Data to be included into file.
        @param type: File type (default: 'w', write mode).
        @return: Returns a tupla: (
            True or False [if an error have occured],
            Error description).
        """
        try:
            fh = open(filename, type)
            fh.write(data)
            fh.close()
            return True,None
        except Exception,e:
            return False,e
            
    @staticmethod        
    def save_binary(filename, data):
        """
        Saves data into a file (binary mode).

        @param filename: Path to file.
        @param data: Data to be included into file.
        @return: Returns a tupla: (
            True or False [if an error have occured],
            Error description).
        """
        try:
            fh = open(filename, 'wb')
            fh.write(data)
            fh.close()
            return True,None
        except Exception,e:
            fh and fh.close()
            return False,e
            
    @staticmethod        
    def read(filename):
        if File.exists(filename) is False:
            File.touch(filename)
        return File.load(filename)
            
    @staticmethod        
    def readlines(filename):
        if File.exists(filename) is False:
            return None
        try:
            fh = open(filename, 'r')
            data = fh.readlines()
            fh.close()
            return data
        except Exception,e:
            fh and fh.close()
            return None
     
    @staticmethod   
    def load(filename):
        """
        Returns data from a file.

        Keyword arguments:
        filename -- file name (path)
        """
        try:
            fh = open(filename)
            data = fh.read()
            fh.close()
        except IOError, (errno, errstr):
            #print("I/O Error (%s): %s <%s>" % (errno, errstr, filename))
            data = None
        return data
    
    @staticmethod        
    def touch(filename):
        """
        Touches a file. Creates an empty file.

        Keyword arguments:
        filename -- file name (path)
        """
        open(filename, 'w').close() 
    
    @staticmethod        
    def copy(origin, destiny):
        """
        Copies a file in another file.

        Keyword arguments:
        origin -- file name (path) to copy
        destiny -- file name (path)
        """
        import shutil
        shutil.copy2(origin, destiny)
    
    @staticmethod        
    def remove(filepath):
        """
        Deletes a file.

        Keyword arguments:
        filename -- file name (path)
        """
        if File.exists(filepath):
            os.remove(filepath)
        
    @staticmethod
    def make_dir(dir, user=None, perms=None):
        """ 
        Create a folder
        """
        if File.check_dir(dir) is False:
            # Checks if folder structure exists
            parent_dir=os.path.dirname(dir)
            if File.check_dir(parent_dir) is False: 
                File.make_dir(parent_dir,user,perms)
                
            if perms is None:
                perms = File.perms
            else:
                perms = int(perms, 8)
            
            os.umask(0)
            os.mkdir(dir, perms)
            if user is not None:
                import pwd
                user_own = -1 #pwd.getpwnam(user).pw_uid
                user_grp = pwd.getpwnam(user).pw_gid
                os.chown(dir, user_own, user_grp)
        
    @staticmethod
    def remove_dir(dirname):
        """ 
        Delete a Folder
        """
        if os.path.isdir(dirname):
            for root, dirs, files in os.walk(dirname, topdown=False):
                # borramos los archivos
                for name in files:
                    os.remove(os.path.join(root, name))
                # borramos los subdirectorios
                for name in dirs:
                    os.rmdir(os.path.join(root, name))
            # ahora podemos borrar el directorio
            os.rmdir(dirname)
            
    @staticmethod
    def clear_dir(dirname):
        """ 
        Delete a Folder
        """
        if os.path.isdir(dirname):
            for root, dirs, files in os.walk(dirname):
                for f in files:
                    os.unlink(os.path.join(root, f))
                for d in dirs:
                    shutil.rmtree(os.path.join(root, d))
                    
    @staticmethod        
    def copy_dir(origin, destiny):
        """
        Copyes a file in another file.

        Keyword arguments:
        origin -- file name (path) to copy
        destiny -- file name (path)
        """
        import shutil
        shutil.copytree(origin, destiny)
        
    @staticmethod
    def file_size(filepath):
        try:
            return os.path.getsize(filepath)
        except:
            return 0
    
    @staticmethod
    def file_version_by_size(filepath, size=1048576):
        if File.file_size(filepath) > size:
            destiny = '%s_%i' % (filepath, int(time.time()))
            os.rename(filepath, destiny)
            File.touch(filepath)
            
            