# -*- coding: utf-8 -*- 

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os.path

from mil.base import File, YAMLFile
from mil.handler import Handler,isNotHandler,yamlHandler
from mil.base import Storage

class Configuration:
    """
    This class implements a configurable set of options
    for use in anything that needs settings that
    are to be stored in a YAML formated file.
    """
    def __init__(self, file_constants=None, handler=None):
        """
        Initialize configure class.
        """
        self.__constants = Storage()
        self.__setting = Storage()
        self.handler = None
        
        if file_constants is not None:
            data = YAMLFile.load(file_constants)
            self.__constants = Storage(data)
        
        if handler is not None:
            # Checks if handler is a Handler instance
            if isinstance(handler, Handler):
                self.handler = handler
            else:
                raise isNotHandler(str(handler))
            
            # Load data from handler
            self.__setting = self.handler.load()

   
    def settings(self, data=None):
        """
        Returns the value of a settings object

        Keyword arguments:
        data -- initial data (dict or Storage)
        """
        if data is None:
            return self.__setting
        else:
            self.__setting = Storage(data)
            
    def __repr__(self):
        return "<Configuration: %s %s >" % (self.__constants, self.__setting)
    
    def __get_all_data(self):
        return dict(self.__constants.items() + self.__setting.items())
                      
    def get(self, *args):
        """
        Returns the value of a settings object. (Returns None if setting not exists)

        Keyword arguments:
        args -- keys values
        """
        try:
            ret_val = self.__get_all_data()[args[0]]
            for arg in args[1:]:
                ret_val = ret_val[arg]
            return ret_val
        except:
            return None
                      
    def get_all(self):
        """
        Returns all values of a settings object. 
        """
        return self.__get_all_data()
           
    def set(self, *args):
        """
        Returns the value of a settings object. (Returns None if setting not exists)

        Keyword arguments:
        args -- keys values
        """        
        if len(args) > 2:
            try:
                key=self.__setting[args[0]]
            except:
                key={}
                self.__setting[args[0]]=key
            for arg in args[1:-2]:
                try:
                    key=key[arg]
                except:
                    key[arg]={}
                    key=key[arg]
                    
            key[args[-2]]=args[-1]
        else:
            self.__setting[args[0]]=args[1]
        
        if self.handler:
            self.handler.save(self.__setting)
        
    def has(self, *args):
        """
        Returns the value of a settings object. (Returns None if setting not exists)

        Keyword arguments:
        name -- setting name
        """
        try:
            return True if self.get(*args) else False
        except:
            return False

def file_configuration(constants_path, file_path, skel_path=None):
    def prepare_config_file(file_path, skel_base=None):
        if os.path.exists(file_path) is False:
            # Checks filesystem
            File.make_dir(os.path.dirname(file_path))
            # Prepare configuration file
            if skel_base is None:
                File.touch(file_path)
            else:
                skel_path = os.path.join(skel_base, os.path.basename(file_path))
                File.copy(skel_path, file_path)
    # Prepare files
    prepare_config_file(constants_path, skel_path)
    prepare_config_file(file_path, skel_path)
    # Get file handler
    _handler=yamlHandler(filepath=file_path)
    return Configuration(file_constants=constants_path, handler=_handler)

        