# -*- coding: utf-8 -*-

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import re
import json
from collections import deque

from mil.base import Storage
from mil.base import URL

from smobject import SMObject
from smdatastructure import SMDataStructure

class SMComponentNode(SMObject):
    """Base class for state machine workflow components."""
    type = 'Base'
    def __init__(self, param=None, globals={}):
        self.connections = {}
        self.connections['output'] = None
        self.param = param
        self.globals = globals
        self.data = None
        self.output = 'output'
        self.error = None
        
        self.next_step = False
        self.url = None
        
        self.event = False
        self.status = False
        
    def __str__(self):
        return '{param : "%s", conns : %s}' % (self.param, self.connections)

    def __call__(self, values):
        """
        It must return a tuple: component, data
        """ 
        if isinstance(values, SMDataStructure):
            self.data = values.data
        else:
            self.data = values
            
        self._prepare_parameters()
        if self._sensor():
            self._control()
            self._working()
        
        return self._return()

    def connect(self, component, name='output'):
        self.connections[name] = component
        
    def set_param(self, param):
        self.param = param
        
    def set_output(self, name):
        self.output = name

    def get_connection(self, name=None):
        if name is None:
            name = self.output
        return self.connections[name]

    def output_ports(self):
        return sorted(self.connections.keys())
    
    def set_error(self, message):
        self.error = message
#        print 'ERROR: %s' % message

    def _prepare_environment(self):
        globals_data = globals()
        globals_data['current'] = self.globals
        for key, value in self.globals.globalenv.iteritems():
            globals_data[key] = value
        return globals_data
    
    def _prepare_parameters(self):
        if len(self.param) > 0:
            # Split into ĺines
            lines = self.param.split('\n')
            for line in lines:
                (param, value) = line.split(':')
                if param == 'url':
                    # NOTE: URL() parameter f must be a string, it can not be 
                    # an unicode string
                    if value.count(','):
                        elems = value.split(',')
                    else:
                        elems = [value, 'index']
                        
                    # Data parameters
                    if self.data is not None:
                        if len(elems) == 3:
                            if isinstance(self.data, dict):
                                params_flow = self.data
                            else:
                                params_flow = [self.data]
                            elems[2] = "%s/%s" % (elems[2],  
                                                  "/".join(params_flow))
                        else:
                            if isinstance(self.data, dict):
                                params_flow = self.data
                            else:
                                params_flow = [self.data]
                            elems.append("/".join(params_flow))
                        
                    if len(elems) == 3:
                        controler = str(elems[0].strip())
                        function = str(elems[1].strip())
                        args_l = []
                        for arg in elems[2].strip().split('/'):
                            args_l.append(str(arg))
                        self.url = URL(r=self.globals.request,
                                       c=controler,
                                       f=function,
                                       args=args_l)
                    else:
                        controler = str(elems[0].strip())
                        function = str(elems[1].strip())
                        self.url = URL(r=self.globals.request,
                                       c=controler,
                                       f=function)
                elif param == 'event':
                    if value.strip().lower() == 'false':
                        self.event = False
                    else:
                        self.event = True
                elif param == 'action': 
                    expr = self._process_action(value.strip())
                    if expr is not None:
                        self.status = eval(expr, self._prepare_environment())
                elif param == 'cond': 
                    expr = None
                    actions = re.split('\{|\}', value.strip())
                    if len(actions) == 1:
                        expr = self._process_action(actions[0])
                    else:
                        elem_l = []
                        for action in actions:
                            value = action.strip()
                            if len(value) > 0:
                                value = self._process_action(value)
                                elem_l.append(value)
                        expr = " ".join(elem_l)
                        
                    if expr is not None:
                        self.status = eval(expr, self._prepare_environment())
                        
    def _process_action(self, value):   
        expr = None
        if value.count(',') == 2:
            elems = value.split(',')
            if elems[1] == 'ne':
                operator = '!='
            if elems[1] == 'is':
                operator = 'is'
            else:
                operator = '=='
            expr = "current.%s %s %s" % (elems[0].strip(), 
                                          operator, 
                                          elems[2].strip())
        else:
            value = value.strip()
            if value in ['and', 'or']:
                expr = value
            else:
                expr = "current.%s" % (value)        
        
        return expr 
        
    def _control(self):
        return
    
    def _working(self):
        self.set_output('output')

    def _return(self):
        return SMDataStructure(connection = self.get_connection(),
                               data = self.data,
                               next_step = self.next_step,
                               called_page = self.url,
                               error = self.error)
   
class SMOpenNode(SMComponentNode):    
    def _sensor(self):
        if not 'output' in self.connections.keys():
            self.set_error('You must define output Component')
            return False
        self.next_step = True
        return True
            
class SMConnectorNode(SMComponentNode):    
    def _sensor(self):
        if not 'con_out1' in self.connections.keys():
            self.set_error('You must define output Component')
            return False
        return True
    
    def _control(self):
        if self.param is not None: 
            self.status = True
        else:
            self.status = False
        
    def _working(self):
        if self.status:
            self.set_output('con_out1')
            self.next_step = True
        else:
            self.set_output(None)

class SMBooleanNode(SMComponentNode):    
    def _sensor(self):
        if not 'con_outt' in self.connections.keys():
            self.set_error('You must define True Component')
            return False
        if not 'con_outf' in self.connections.keys():
            self.set_error('You must define False Component')
            return False
        self.next_step = True
        return True
    
    def _control(self):
        if self.param is None:
             self.status = False
    
    def _working(self):
        if self.status:
            self.set_output('con_outt')
        else:
            self.set_output('con_outf')

class SMEndNode(SMComponentNode):  
    def _sensor(self):
        if self.param is None:
            return False  
        else:
            return True
        
    def _control(self):
        self.status = True
                
#    def _working(self):
#        if self.url:
#            redirect(self.url)     
    
#class SMPagecallNode(SMComponentNode):
#    def _sensor(self):
#        return True
#        
#    def _control(self):
#        self.url = self.param
#        
#    def _working(self):
#        if self.url:
#            redirect(self.url)     
    
#class SMPageviewNode(SMComponentNode):
#    def _sensor(self):
#        return True     
