# -*- coding: utf-8 -*-

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import re
import json
from collections import deque

from mil.base import Storage
from mil.base import URL

from smobject import SMObject
from smdatastructure import SMDataStructure

class SMComponentNode(SMObject):
    """Base class for state machine workflow components."""
    type = 'Base'
    def __init__(self, param=None, globals={}):
        self.connections = {}
        self.connections['output'] = None
        self.param = param
        self.globals = globals
        self.data = None
        self.output = 'output'
        self.error = None
        
        self.next_step = False
        self.url = None
        
        self.event = False
        self.status = False
        
    def __str__(self):
        return '{param : "%s", conns : %s}' % (self.param, self.connections)

    def __call__(self, values):
        """
        It must return a tuple: component, data
        """ 
        if isinstance(values, SMDataStructure):
            self.data = values.data
        else:
            self.data = values
        
        self._prepare_parameters()
        if self._sensor():
            self._control()
            self._working()
        
        return self._return()

    def connect(self, component, name='output'):
        self.connections[name] = component
        
    def set_param(self, param):
        self.param = param
        
    def set_output(self, name):
        self.output = name

    def get_connection(self, name=None):
        if name is None:
            name = self.output
        return self.connections[name]

    def output_ports(self):
        return sorted(self.connections.keys())
    
    def set_error(self, message):
        self.error = message
#        print 'ERROR: %s' % message

    def _prepare_environment(self):
        try:
            globals_data = globals()
            globals_data['current'] = self.globals
            try:
                for key, value in globals().iteritems():
                    globals_data[key] = value
            except:
                pass
            return globals_data
        except Exception, e:
            raise Exception('SMComponentNode._prepare_environment: %s' % str(e))
    
    def _prepare_parameters(self):
        try:
            if len(self.param) > 0:
                # Split into ĺines
                lines = self.param.split('\n')
                for line in lines:
                    (param, value) = line.split(':')
                    if param == 'url':
                        # NOTE: URL() parameter f must be a string, it can not be 
                        # an unicode string
                        if value.count(','):
                            elems = value.split(',')
                        else:
                            elems = [value, 'index']
                            
                        # Data parameters
                        if self.data is not None:
                            if len(elems) == 3:
                                if isinstance(self.data, dict):
                                    params_flow = self.data
                                else:
                                    params_flow = [self.data]
                                elems[2] = "%s/%s" % (elems[2],  
                                                      "/".join(params_flow))
                            else:
                                if isinstance(self.data, dict):
                                    params_flow = self.data
                                else:
                                    params_flow = [self.data]
                                elems.append("/".join(params_flow))
                            
                        if len(elems) == 3:
                            controler = str(elems[0].strip())
                            function = str(elems[1].strip())
                            args_l = []
                            for arg in elems[2].strip().split('/'):
                                args_l.append(str(arg))
                            self.url = URL(r=self.globals.request,
                                           c=controler,
                                           f=function,
                                           args=args_l)
                        else:
                            controler = str(elems[0].strip())
                            function = str(elems[1].strip())
                            self.url = URL(r=self.globals.request,
                                           c=controler,
                                           f=function)
                    elif param == 'event':
                        if value.strip().lower() == 'false':
                            self.event = False
                        else:
                            self.event = True
                    elif param == 'action': 
                        expr = self._process_action(value.strip())
                        if expr is not None:
                            self.status = eval(expr, self._prepare_environment())
                    elif param == 'cond': 
                        expr = None
                        actions = re.split('\{|\}', value.strip())
                        if len(actions) == 1:
                            expr = self._process_action(actions[0])
                        else:
                            elem_l = []
                            for action in actions:
                                value = action.strip()
                                if len(value) > 0:
                                    value = self._process_action(value)
                                    elem_l.append(value)
                            expr = " ".join(elem_l)
                            
                        if expr is not None:
                            self.status = eval(expr, self._prepare_environment())
        except Exception, e:
            raise Exception('SMComponentNode.__prepare_paramaters__: %s' % str(e))
                        
    def _process_action(self, value):   
        expr = None
        if value.count(',') == 2:
            elems = value.split(',')
            if elems[1] == 'ne':
                operator = '!='
            if elems[1] == 'is':
                operator = 'is'
            else:
                operator = '=='
            expr = "current.%s %s %s" % (elems[0].strip(), 
                                          operator, 
                                          elems[2].strip())
        else:
            value = value.strip()
            if value in ['and', 'or']:
                expr = value
            else:
                expr = "current.%s" % (value)        
        
        return expr 
        
    def _control(self):
        return
    
    def _working(self):
        self.set_output('output')

    def _return(self):
        return SMDataStructure(connection = self.get_connection(),
                               data = self.data,
                               next_step = self.next_step,
                               called_page = self.url,
                               error = self.error)
   
class SMOpenNode(SMComponentNode):    
    def _sensor(self):
        if not 'output' in self.connections.keys():
            self.set_error('You must define output Component')
            return False
        self.next_step = True
        return True
            
class SMConnectorNode(SMComponentNode):    
    def _sensor(self):
        if not 'con_out1' in self.connections.keys():
            self.set_error('You must define output Component')
            return False
        return True
    
    def _control(self):
        if self.param is not None: 
            self.status = True
        else:
            self.status = False
        
    def _working(self):
        if self.status:
            self.set_output('con_out1')
            self.next_step = True
        else:
            self.set_output(None)

class SMBooleanNode(SMComponentNode):    
    def _sensor(self):
        if not 'con_outt' in self.connections.keys():
            self.set_error('You must define True Component')
            return False
        if not 'con_outf' in self.connections.keys():
            self.set_error('You must define False Component')
            return False
        self.next_step = True
        return True
    
    def _control(self):
        if self.param is None:
             self.status = False
    
    def _working(self):
        if self.status:
            self.set_output('con_outt')
        else:
            self.set_output('con_outf')

class SMEndNode(SMComponentNode):  
    def _sensor(self):
        if self.param is None:
            return False  
        else:
            return True
        
    def _control(self):
        self.status = True
                
#    def _working(self):
#        if self.url:
#            redirect(self.url)     
    
#class SMPagecallNode(SMComponentNode):
#    def _sensor(self):
#        return True
#        
#    def _control(self):
#        self.url = self.param
#        
#    def _working(self):
#        if self.url:
#            redirect(self.url)     
    
#class SMPageviewNode(SMComponentNode):
#    def _sensor(self):
#        return True   

class SMWorkFlow(object):
    def __init__(self, workflow, globals={}):
        self.workflow = workflow
        self.nodes = {}
        self.connections = {}
        self.globals = globals
        
        self._initialize()
        self.first_node = self._find_first_component()

    def __str__(self):
        data = {'nodes': self.nodes, 'connections' : self.connections, 'first': self.first_node}
        return json.dumps(data, indent=4)

    __repr__ = __str__
    
    def _initialize(self):
        if self.workflow is not None:
            data = json.loads(self.workflow)

            for node in data:
                tmp = Storage()
                # type
                tmp.type = data[node]['type']
                tmp.param = data[node]['param']
    
                self.nodes[node] = tmp
                
                # Connections    
                node_conns = data[node]['connector']
                
                source = node
                for conn in node_conns['output']:
                    output = node_conns['output'][conn]['connection']
                
                    if output != '':
                        destination, d_port = output .split(".", 2)
                        if not source in self.connections:
                            self.connections[source] = {}
                        if not destination in self.connections:
                            self.connections[destination] = {}
        
                        mappings = self.connections[source]
                        mappings[conn] = destination
                    
    def _find_first_component(self):
        notFirst = []
        names = self.connections.keys()
        for name in names:
            mappings = self.connections[name]
            ports = mappings.keys()
            for port in ports:
                destination = mappings[port]
                if destination != None:
                    if not destination in notFirst:
                        notFirst.append(destination)
        first_comp = [name for name in names if not name in notFirst]
        if len(first_comp) == 1:
            return first_comp[0]
        raise Exception("Workflow does not have a clear first component")    
        
    def _prepare(self, string):
        self.connections.clear()
        
        lines = string.split("\n")
        for line in lines:
            if line.find("=>") == -1:
                raise Exception('Bad formed workflow')
            
            line = line.replace(' ','')
            source, destination = line.split("=>", 2)

            port_index = source.find(".")
            if port_index > 0:
                port = source[port_index+1:]
                source = source[0:port_index]
            else:
                port = 'output'

            if not source in self.connections:
                self.connections[source] = {}
            if not destination in self.connections:
                self.connections[destination] = {}

            mappings = self.connections[source]
            mappings[port] = destination
        return True
    
    def run(self, values=None, first_component=None):
        if first_component is None:
            first_component = self.first_node
                      
        q = deque([(first_component, values)])    
        while len(q) > 0:
            component, values = q.popleft()
            #print 'Qeue: %s' % component
            if component is not None:
                returned = self.step(component, values)
                q.append((returned.connection, returned.data))
        return values
    
    def step(self, name, values): 
        """
        next_component, next_values
        """
        try:
            component = self.prepare_step(name)
            return component(values)
        except Exception, e:
            raise Exception ('SMWorkFlow.step: %s' % str(e))
    
    def prepare_step(self, name):
        node = self.nodes[name]
        try:
            class_name = "SM%sNode" % node['type'].capitalize()  
            component = globals()[class_name](node['param'], self.globals)
            
            for conn in self.connections[name]:
                component.connect(self.connections[name][conn], conn)
            return component
        except Exception, e:
            raise Exception('SMWorkFlow.prepare_step %s' % str(e))
        
    
#    def prepare_step(self, name):
#        node = self.nodes[name]
#        try:
#            class_name = "SM%sNode" % node['type'].capitalize()  
#            
#            # Use imported globals to recognize the python path
#            rel_globals = self.globals.globalenv
#            app_name = rel_globals['current'].request.application
#            local_import_aux = rel_globals['local_import']
#            module = local_import_aux('mil.statemachine.smwfelements',
#                                      app=app_name)
#            inst = getattr(module, class_name)
#            component = inst(node['param'], self.globals)
#            
#            for conn in self.connections[name]:
#                component.connect(self.connections[name][conn], conn)
#            return component
#        except Exception, e:
#            raise Exception('SMWorkFlow.prepare_step %s' % str(e))
        
    