# -*- coding: utf-8 -*-

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

from collections import deque

from smflow import SMFlow
from smdatastructure import SMDataStructure

class SMEngine(object):
    """State Machine
    """
    def __init__(self, flow, path):
        if isinstance(flow, SMFlow):
            self.flow = flow
        else:
            raise TypeError('flow must be SMFlow instance')

        if (path is None) or (len(path) == 0):
            raise ValueError('path must not be empty')
        else:
            self.path = path
        
        self.components = self._instantiate()            

    def __str__(self):
        return '{flow : %s, path : %s}' % (self.flow.get_name(), self.path)

    __repr__ = __str__
    
    def identifier(self):
        return '%s-%s' % (self.flow.get_name(), self.path)
    
    def components_path(self, path):
        self.path = path      
        
    def get_component(self, name):
        return self.components[name]
    
    def _instantiate(self):
       components = {}

       # load all components mentioned in workflow into memory
       # if component is a factor, retrieve the components
       # selected for this factor and load them as well
       names = self.flow.connections.keys()
       for name in names:
           components[name] = self.load_component(name, self.path)

       for name in names:
           comp  = components[name]
           connections = self.flow.connections[name]
           ports = connections.keys()
           for port in ports:
               target = connections[port]
               tcomp = components[target]
               comp.connect(tcomp, port)

       return components

    def _find_first_component(self):
        notFirst = []
        names = self.flow.connections.keys()
        for name in names:
            mappings = self.flow.connections[name]
            ports = mappings.keys()
            for port in ports:
                destination = mappings[port]
                if destination != None:
                    if not destination in notFirst:
                        notFirst.append(destination)
        first_comp = [name for name in names if not name in notFirst]
        if len(first_comp) == 1:
            return first_comp[0]
        raise Exception("Workflow does not have a clear first component")
        
    def first_component(self):
        return self.components[self._find_first_component()]  
       
    def first_component_name(self):
        return self._find_first_component() 
    
    def next_component_name(self, current):
        try:
            return self.flow.connections[current]['output']
        except KeyError:
            return None

    def run(self, values=None, first_component=None):
        if first_component is None:
            first_component = self.first_component()
                      
        q = deque([(first_component, values)])    
        while len(q) > 0:
            component, values = q.popleft()
#            print 'Qeue: %s' % component
            if component is not None:
                returned = self.step(component, values)
                q.append((returned.connection, returned.data))
        return values
    
    def step(self, component, values): 
        """
        next_component, next_values
        """
        return component(values)
        

    def load_component(self, name, path=None):
        if path is None:
            path = self.path
        module = self._import_aux(path)
        comp = getattr(module, name)
        instance = comp(self)
        return instance

    def _import_aux(self, name, force=False, path=None):
        """
        In apps, instead of importing a local module
        (in applications/app/modules) with::
    
           import a.b.c as d
    
        you should do::
    
           d = local_import('a.b.c')
    
        or (to force a reload):
    
           d = local_import('a.b.c', reload=True)
    
        This prevents conflict between applications and un-necessary execs.
        It can be used to import any module, including regular Python modules.
        """
        items = name.replace('/','.')
        if path is not None:
            name = "%s.%s" % (path, items)
        else:
            name = items
        module = __import__(name)
        for item in name.split(".")[1:]:
            module = getattr(module, item)
        if force:
            reload(module)
        return module
