# -*- coding: utf-8 -*- 

from decimal import *

def moneyfmt(value, places=2, curr='', sep=',', dp='.', pos='', neg='-', trailneg=''):
    """Convert Decimal to a money formatted string.

    places:  required number of places after the decimal point
    curr:    optional currency symbol before the sign (may be blank)
    sep:     optional grouping separator (comma, period, space, or blank)
    dp:      decimal point indicator (comma or period)
             only specify as blank when places is zero
    pos:     optional sign for positive numbers: '+', space or blank
    neg:     optional sign for negative numbers: '-', '(', space or blank
    trailneg:optional trailing minus indicator:  '-', ')', space or blank

    >>> d = Decimal('-1234567.8901')
    >>> moneyfmt(d, curr='$')
    '-$1,234,567.89'
    >>> moneyfmt(d, places=0, sep='.', dp='', neg='', trailneg='-')
    '1.234.568-'
    >>> moneyfmt(d, curr='$', neg='(', trailneg=')')
    '($1,234,567.89)'
    >>> moneyfmt(Decimal(123456789), sep=' ')
    '123 456 789.00'
    >>> moneyfmt(Decimal('-0.02'), neg='<', trailneg='>')
    '<0.02>'

    """
    q = Decimal(10) ** -places      # 2 places --> '0.01'
    sign, digits, exp = value.quantize(q).as_tuple()
    result = []
    digits = map(str, digits)
    build, next = result.append, digits.pop
    if sign:
        build(trailneg)
    for i in range(places):
        build(next() if digits else '0')
    build(dp)
    if not digits:
        build('0')
    i = 0
    while digits:
        build(next())
        i += 1
        if i == 3 and digits:
            i = 0
            build(sep)
    build(curr)
    build(neg if sign else pos)
    return ''.join(reversed(result))

def format_money(value):
    value = Decimal(value)
    return moneyfmt(value, places=0, sep='.', dp='')
    #return '{:17,.0f}'.format(int(value)).replace(',','.')

def format_text_to_view(value):
    if value=='' or value is None:
        return cfgHnd.get('empty_text')
    else:
        return XML(value.replace('\n','<br />'),sanitize=True)

def format_number_to_view(value):
    if value is None:
        return cfgHnd.get('empty_text')
    else:
        return str(value)
    
def format_import_to_view(value):
    if value is None:
        return cfgHnd.get('empty_text')
    else:
        value=Decimal(value)
        return moneyfmt(value, places=0, sep='.', dp='')

def format_text_to_rtf(value):
    if value=='' or value is None:
        return cfgHnd.get('empty_text')
    else:
        try:
            udata=value.decode('utf-8')
            return udata.encode('iso-8859-1','ignore')
        except:
            return value
    
def format_import_to_rtf(value):
    if value is None:
        return cfgHnd.get('empty_text')
    else:
        value=Decimal(value)
        return moneyfmt(value, places=2, sep='.', dp=',').strip()
    
def format_import_no_decimals_to_rtf(value):
    if value is None:
        return cfgHnd.get('empty_text')
    else:
        value=Decimal(value)
        return moneyfmt(value, places=0, sep='.', dp='').strip()

def truncate_to(value,length):
    if (length is None) or (length == 0):
        return value
    if len(value) > length:
        return value[0:length]
    else:
        return value
    
def remove_special_chars(value):
#    value.replace("\r\n", ' ') 
#    value.replace("\n", ' ')  
#    value.replace("\t", ' ')  
    value = value.replace("\"", '""')  
    value = value.replace("#", '') 
    return value


