# -*- coding: utf-8 -*-

# Copyright Daniel Gonzalez     <dgzaballos@gmail.com>
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import datetime
import os

from gluon import *
from gluon.storage import Storage

from mil.base import File
from mil.tools import get_unique_id, get_file_id
from mil.tools.common import return2json, escape2html
from masters.file import FileModel

from project.models.element import ElementModel
from project.models.tag import TagModel
from project.models.comment import CommentModel

class Element():
    def __init__(self, identifier):
        self.identifier = identifier
        self.db = current.db
        
        # Data Models
        FileModel.define_table(self.db)
        ElementModel.define_table(self.db)
        TagModel.define_table(self.db)
        CommentModel.define_table(self.db)
        
        self.query = None
        
    def default_query(self,type):
        return (self.db.element_has_element.type==type) & (self.db.element_has_element.parent==self.identifier)
                
    def set_query(self,query):
        self.query=query
    
    def get_children_from_db(self,limit=None):
        if limit is not None:
            children=self.db(self.query).select(self.db.element_has_element.id,
                                           self.db.element_has_element.numeral,
                                           orderby=self.db.element_has_element.sortable,
                                           limitby=limit)
        else:
            children=self.db(self.query).select(self.db.element_has_element.id,
                                           self.db.element_has_element.numeral,
                                           orderby=self.db.element_has_element.sortable)
        return children
        
    def get_children(self,child=None,limit=None,level=0):
        children=self.get_children_from_db(child,limit)
       
        child_list=[]
        if children is not None:
            for child in children:
                child_list.append(self.prepare_child(child,level))
        return child_list
    
    def prepare_child(self,child,level=0):
        return Storage(id=child.id, numeral=child.numeral, child=child.child)
        
    def insert_child(self,child,type):
        # Join child to specific element
        query=(self.db.element_has_element.type==type) & (self.db.element_has_element.parent==self.identifier)
        try:
            max_id = self.db(query).select(self.db.element_has_element.sortable.max())[0]['_extra']['MAX(element_has_element.sortable)']
            if max_id is None:
                max_id=0
        except KeyError:
            max_id=0
        try:
            numeral = self.db(query).select(self.db.element_has_element.numeral.max())[0]['_extra']['MAX(element_has_element.numeral)']
            if numeral is None:
                numeral=0
        except KeyError:
            numeral=0
        max = int(max_id) + 1
        num = int(numeral) + 1
        id=self.db.element_has_element.insert(parent=self.identifier,
                                         child=child,
                                         type=type,
                                         numeral=num,
                                         sortable=max)
        return (id,num)
    
    def delete_child(self,child,type):
        # Delete child to specific element relation
        query=((self.db.element_has_element.type==type) & (self.db.element_has_element.parent==self.identifier) & (self.db.element_has_element.child==child))
        
        return self.db(query).delete()
    
    def sort(self,vars,type,sortable_field='sortable'): 
        for i,id in enumerate(vars.order.split(',')):
            self.db((self.db.element_has_element.id==id)&(self.db.element_has_element.type==type)&(self.db.element_has_element.parent==self.identifier)).update(**{sortable_field:i+1})
    
    ## TAG    
    def get_tags_row(self,identifier):
        tags = self.db(self.db.tag.element==identifier).select(self.db.tag.id,self.db.tag.value)
        tag_list=[]
        if tags is not None:
            for tag in tags:
                tag_list.append({'id':tag.id, 'value':tag.value})
        return tag_list
    
    def add_tag(self,pid,tid,id):        
        # Form
        from gluon.sql import SQLField
        from gluon.sqlhtml import form_factory, IS_NOT_EMPTY
        form=form_factory(SQLField('tag',requires=IS_NOT_EMPTY()))
    
        bol_ok=True
        if form.accepts(current.request.vars, current.session):
            n_tag=current.request.vars.tag
            
            tmp = []
            if n_tag != '':
                tmp=self.add_tag_list(n_tag)
    
            params={}
            params['div_id']='mm_add'
            params['data']={'tid':id,'tags':tmp}
            bol_ok=False
            import gluon.contrib.simplejson
            from gluon import XML
            form=XML(gluon.contrib.simplejson.dumps(params))
            
        return dict(form=form,show=bol_ok)
    
    def add_tag_list(self,tags):
        # Split tags by comma
        import csv
        reader=csv.reader([tags], skipinitialspace=True)
        tmp=[]
        for row in reader:
            for e_tag in row:     
                # Create Tag
                tag_id=self.insert_tag(self.identifier,e_tag)
                tmp.append({'id':tag_id ,'value': e_tag})
        return tmp
    
    def insert_tag(self,tag_identifier,tag_value):
        return self.db.tag.insert(element=tag_identifier,
                             value=tag_value,
                             project=current._user.project.ident)
    
    def delete_tag(self,id):
        # Delete tag
        self.db(self.db.tag.id==id).delete()
        # Return new list to view
        return current.T('Tag deleted')
    
    ## COMMENT    
    def get_comments_row(self,identifier):
        comments = self.db(self.db.comment.element==identifier).select(
                                            self.db.comment.id,
                                            self.db.comment.content,
                                            self.db.comment.author,
                                            self.db.comment.create_date)
        comment_list=[]
        if comments is not None:
            for comment in comments:
                comment_list.append(
                    {'id':comment.id,
                     'content':escape2html(return2json(comment.content)),
                     'author':comment.author,
                     'date': str(comment.create_date)})
        return comment_list
    
    def add_comment(self,pid,tid,id):
        # Form
        from gluon.sql import SQLField
        from gluon.sqlhtml import form_factory, IS_NOT_EMPTY
        form=form_factory(SQLField('comment','text',requires=IS_NOT_EMPTY()))
    
        bol_ok=True
        if form.accepts(current.request.vars, current.session):
            # Prepare data
            comment_text=current.request.vars.comment
            comment_author=current._user.email
            comment_date=datetime.datetime.now()
            
            # Add comment
            comm_id=self.insert_comment(self.identifier,
                                        comment_text,
                                        comment_author,
                                        comment_date)
            
            # Prepare data to view
            comment = {'id': comm_id ,
                       'content': escape2html(return2json(comment_text)),
                       'author': comment_author,
                       'date': comment_date.strftime('%Y-%m-%d %H:%M:%S')}
            params = {}
            params['div_id'] = 'mm_add'
            params['data'] = {'tid': id,'comment': comment}
            bol_ok = False
            import gluon.contrib.simplejson
            from gluon import XML
            form = XML(gluon.contrib.simplejson.dumps(params))
            
        return dict(form=form,show=bol_ok)
    
    def insert_comment(self,identifier,comment_text,comment_author,comment_date):
        return self.db.comment.insert(element=identifier,
                                 content=comment_text,
                                 author=comment_author,
                                 create_date=comment_date,
                                 project=current._user.project.ident)
    
    ## DOCUMENT    
    def get_docs_row(self,identifier):
        query=((self.db.file.identifier==self.db.element_has_element.child)&(self.db.element_has_element.type=='doc')&(self.db.element_has_element.parent==identifier))
        rows = self.db(query).select(self.db.file.identifier,
                                self.db.file.name,
                                self.db.file.creator,
                                self.db.file.created,
                                self.db.element_has_element.numeral)
        row_list=[]
        if rows is not None:
            for row in rows:
                row_list.append({'id':row.file.identifier, 
                                 'name':row.file.name,
                                 'title': "%s - %s"%(row.file.creator,row.file.created),
                                 'url':str(URL(r=current.request, c='file', f='show', 
                                               args=(current._user.project.name,
                                                     row.file.identifier)))})
        return row_list
    
    def add_document(self,tid):    
        if current.request.vars.qqfile!=None:
            try:
                file_name=current.request.vars.qqfile   
                type=file_name.split(".")[-1] or ''
                
                # Save file content     
                file_realname="file.file_content.%s.%s"%(get_file_id(),type)
                file_content = current.request.body.read()
                file_path=os.path.join(current._config.app_path,'uploads',file_realname)
                File.save_binary(file_path, file_content)
                
                # Get file params
                file_params=Storage()
                file_params.filename=file_name
                file_params.type=file_name.split(".")[-1] or ''
                file_params.bufsize=os.path.getsize(file_path)
                file_params.content=file_realname
                
                # File to database
                params=self.add_doc(file_name,file_params) 
                
                params={'success':True,'data':{'id':tid,'params':params}}
            except Exception, e:
                params={'error':e}
        else:
            params={'error': current.T('File not received in server')}
                    
        return params
    
    def add_doc(self,name,file_content):
        # Create Document
        tmp=Storage()
        tmp.identifier=get_unique_id(2,4)
        tmp.name=name
        tmp.project = current._user.project.ident 
        tmp.creator=current._user.email
        tmp.created=datetime.datetime.now()
        tmp.deleted=None
        tmp.file_name=file_content.filename
        tmp.file_format=file_content.type
        tmp.file_size=file_content.bufsize
        tmp.file_content=file_content.content
        id=self.db.file.insert(**tmp)

        self.insert_child(tmp.identifier,'doc')
        param={'id':tmp.identifier,
               'name':tmp.name,
               'title':"%s - %s"%(tmp.creator,tmp.created),
               'url':str(URL(r=current.request, c='file', f='show', 
                             args=(current._user.project.name,
                                   tmp.identifier)))}
        return param
     
    def delete_document(self,id):
        # Get data
        query=((self.db.file.identifier==self.db.element_has_element.child)&(self.db.element_has_element.child==id)&(self.db.element_has_element.type=='doc')&(self.db.element_has_element.parent==self.identifier))
        row = self.db(query).select(self.db.file.identifier,
                               self.db.file.name,
                               self.db.file.file_name,
                               self.db.file.file_content,
                               self.db.element_has_element.numeral).first()
        
        # delete file        
        file_path=os.path.join(current._config.app_path,'uploads',row.file.file_content)
        File.remove(file_path)
        
        # Delete element_has_element
        self.db((self.db.element_has_element.child==id)&(self.db.element_has_element.type=='doc')&(self.db.element_has_element.parent==self.identifier)).delete()
        
        # Delete file
        self.db(self.db.file.identifier==row.file.identifier).delete()
        # Return new list to view
        return current.T('File deleted')
    
    def add_file(self,tid):    
        bol_ok=True            
        return dict(identifier=self.identifier,tid=tid,show=bol_ok)