# -*- coding: utf-8 -*-

# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import types

from gluon import current

from system.authzfile import AuthZFile

class AuthZGroup(object):
    def __init__(self, data): 
        assert type(data) == types.StringType, 'The value must be a string'
        self.data = {} 
        self.__prepare_data(data)
        
    def __prepare_data(self, s_data):
        lines = s_data.split('\n')
        for line in lines:
            if line is not None and line != '':
                tmp = []
                (key, value) = line.split(' = ')
                values = value.split(', ')
                for value in values:
                    tmp.append(value)
                    
                self.data[key] = tmp
            
    def get(self):
        lines = []        
        for item in self.data:
            lines.append('%s = %s' % (item, ', '.join(self.data[item])))
        return '\n'.join(lines)
    
    def set(self, key, value=None):
        # Create an empty group
        if value is None:
            value = ''
        # Create and asiign user
        self.data[key].append(value)
        
    def remove(self, key, item=None):
        # Remove all group
        if item is None:
            if key in self.data.keys():
                del self.data[key]
        
        # Remove a user from group
        if key in self.data.keys():
            if item in self.data[key]:
                self.data[key].remove(item)

class AuthZRepo(object):
    def __init__(self, name, data=None): 
        self.name = name
        self.data = {} 
        if data is not None:
            assert type(data) == types.StringType, 'The value must be a string'
            self.__prepare_data(data)
        
    def __prepare_data(self, s_data):
        if s_data != '':
            lines = s_data.split('\n')
            for line in lines:
                (key, value) = line.split(' = ')
                self.data[key] = value
            
    def get(self):
        lines = []
        
        elems = self.data.keys()
        elems.sort()    # we want to make sure that 'global' repository is always first

        data = map(lambda x : x is None or x, elems)  
        
        for item in data:
            lines.append('%s = %s' % (item, self.data[item]))
        return '\n'.join(lines)
    
    def set(self, key, value):
        if not value in ['rw', 'r']:
            value = 'r'
        self.data[key] = value
        
    def remove(self, item):
        if item in self.data.keys():
            del self.data[item]
            

class AuthZHandler(object):
    def __init__(self, filename):
        self.filename = filename
        self.__initialize()
        
    def __initialize(self):
        self.authz = AuthZFile(self.filename)
        self.authz.read()
        
    def set_group(self, group, user):
        # Add groups and user
        cur_groups = AuthZGroup(self.authz['groups'])
        cur_groups.set(group, user)
        
        # Save file
        self.authz['groups'] = cur_groups.get()
        self.authz.write()
        
    def del_groups(self, group, user=None):
        # Add groups and user
        cur_groups = AuthZGroup(self.authz['groups'])
        cur_groups.remove(group, user)
        
        # Save file
        self.authz['groups'] = cur_groups.get()
        self.authz.write()
        
    def add_repo(self, repo, user=None, perms=None):
        if user is None:
            self.authz[(repo, '/%s' % repo)] = ''
        else:
            if perms is None:
                perms = 'r'
                
            # Get repo
            items = self.authz.repo_keys(repo)
            if len(items) > 0:
                item = items[0]
                c_repo = AuthZRepo(item, self.authz[item])
            
                # Add groups and user
                c_repo.set(user, perms)
                
                # Set data
                self.authz[item] = c_repo.get()
            else: # Create repo                
                c_repo = AuthZRepo(repo)
            
                # Add groups and user
                c_repo.set(user, perms)
                
                # Set data
                self.authz[(repo, '/%s' % repo)] = c_repo.get()
        
        # Save file
        self.authz.write()
        
    def del_repo(self, repo, user=None):  
        # Delete oll repo
        if user is None:
            # Get repo
            items = self.authz.repo_keys(repo)
            for item in items:
                del self.authz[item]
        else:          
            # Get repo
            items = self.authz.repo_keys(repo)
                
            if len(items) > 0:
                item = items[0]
                repo = AuthZRepo(item, self.authz[item])
            
                # Remove groups and user
                repo.remove(user)       
                self.authz[item] = repo.get()
        
        # Save file
        self.authz.write()

            
        