# -*- coding: utf-8 -*- 

# Copyright Daniel Gonzalez     <dgzaballos@gmail.com>
# Copyright Jesus Martinez      <jamarcer@gmail.com>
#
# Este fichero se distribuye bajo la licencia GPL según las
# condiciones que figuran en el fichero 'LICENSE' que se acompaña.
# Si se distribuyera este fichero individualmente, DEBE incluirse aquí
# las condiciones expresadas allí.
#
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###############################################################################
# Please try to conform to PEP-8 as much as possible
# http://www.python.org/dev/peps/pep-0008/
###############################################################################

import os

from mil.tools.regexp import check
from mil.repos.subapp import SubApp
from mil.repos.apache import Apache

from masters.actions import Actions

class SubAppConfigurator:
    def __init__(self, config, db=None, app_path=None, syslog=None):
        self.config = config
        self.syslog = syslog
        if db is None:
            raise Exception('parameter "db" must be defined')
        else:
            self.db = db
        if app_path is None:
            raise Exception('parameter "app_path" must be defined')
        else:
            self.app_path = app_path
        self.mapping = {'11': self._create,         # Create subapplication
                        '12': self._delete,         # Delete subapplication
                        '13': self._join_member,    # Join member to subapp
                        '14': self._unjoin_member,  # Unjoin member to subapp
                        '15': self._add_service,    # Add service to subapp
                        '16': self._remove_service, # Remove service to subapp
                        '17': self._private,        # Subapplication is private
                        '18': self._public}         # Subapplication is public
        self.restart_apache = False
        
    def run(self):
        ap_inst = Actions(self.db)
        actions = ap_inst.get_pending()
        
        for action in actions:
            if str(action['type']) in self.mapping.keys():
                (ok, msg) = self.mapping[str(action['type'])](action['params'])
                ap_inst.update(action['id'], ok, msg)
            else:
                self._log('Unhandled action: %s' % str(action['type']), 
                          'error') 
        
        if self.restart_apache:
            apache = Apache()
            try:
                apache.reload()
            except Exception, e:
                self._log(str(e))
        
    def _create(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        params['user'] = check('EMAIL', data[2])
        params['mail'] = check('EMAIL', data[3])
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.create()
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _delete(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.delete()
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _join_member(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        member = check('EMAIL', data[2]) or None
        if member is None:
            return (False, 'member not specified')
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.add_member(member)
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _unjoin_member(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        member = check('EMAIL', data[2]) or None
        if member is None:
            return (False, 'member not specified')
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.del_member(member)
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _add_service(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        service = check('Name', data[2]) or None
        if service is None:
            return (False, 'service not specified')
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.add_service(service.upper())
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _remove_service(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        service = check('Name', data[2]) or None
        if service is None:
            return (False, 'service not specified')
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.del_service(service.upper())
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _private(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        service = check('Name', data[2]) or None
        if service is None:
            return (False, 'service not specified')
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.auth_service(service.upper(), 'private')
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    def _public(self, data):
        type = data[0] or 'project'
        
        params = {}
        params['name'] = check('NAME', data[1])
        
        service = check('Name', data[2]) or None
        if service is None:
            return (False, 'service not specified')
        
        acl = self.config['acl']
        
        config = {'app': self.config[type],
                  'services': self.config['services'],
                  'private_path': os.path.join(self.app_path, 
                                               'private', 
                                               'libforge_tmpl')
                  }
        try:
            sapp_inst = SubApp(params=params, acl=acl, config=config)
            msg = sapp_inst.auth_service(service.upper(), 'public')
        except Exception,e:
            msg = str(e)
        if msg:
            self._log(msg, 'error') 
            return (False, msg)
        else:
            self.restart_apache = True
        return (True, '')
        
    
    def _log(self, message, level='warning'):
        if self.syslog is not None:
            self.syslog.error('[libforge] %s' % message, level)